__all__ = ("get_oorb_data_dir", "PyOrbEphemerides")

import os
import time
import warnings
from itertools import repeat

import numpy as np
import pandas as pd
import pyoorb as oo


def dtime(time_prev):
    return (time.time() - time_prev, time.time())


def get_oorb_data_dir():
    """Find where the oorb files should be installed"""
    data_path = os.getenv("OORB_DATA")
    if data_path is None:
        # See if we are in a conda enviroment and can find it
        conda_dir = os.getenv("CONDA_PREFIX")
        if conda_dir is not None:
            data_path = os.path.join(conda_dir, "share/openorb")
            if not os.path.isdir(data_path):
                data_path = None
        os.environ["OORB_DATA"] = data_path
    if data_path is None:
        warnings.warn(
            "Failed to find path for oorb data files. "
            "No $OORB_DATA environment variable set, and they are not the usual conda spot"
        )
    return data_path


class PyOrbEphemerides:
    """Generate ephemerides and propagate orbits using the python interface to Oorb.

    Typical usage:
    pyephs = PyOrbEphemerides()
    # Set the orbital parameters, using an lsst.sims.movingObjects.Orbits object
    pyephs.setOrbits(orbits)
    # Generate ephemerides at times 'times'.
    ephs = pyephs.generateEphemerides(times, timeScale='UTC', obscode='I11')

    This class handles the packing and unpacking of the fortran style arrays that
    pyoorb uses, to and from more user-friendly pandas arrays.

    Parameters
    ----------
    ephfile : `str`, optional
        Planetary ephemerides file for Oorb (i.e. de430 or de405).
        Default $OORB_DATA/de430.dat  ($OORB_DATA = $OORB_DIR/data).
    """

    def __init__(self, ephfile=None):
        # Set translation from timescale to OpenOrb numerical representation.
        # Note all orbits are assumed to be in TT timescale.
        # Also, all dates are expected to be in MJD.
        self.time_scales = {"UTC": 1, "UT1": 2, "TT": 3, "TAI": 4}
        self.elem_type = {"CAR": 1, "COM": 2, "KEP": 3, "DEL": 4, "EQX": 5}

        # Set up oorb. Call this once.
        if ephfile is None:
            ephfile = os.path.join(get_oorb_data_dir(), "de430.dat")
        self.ephfile = ephfile
        self._init_oorb()
        self.oorb_elem = None
        self.orb_format = None

    def _init_oorb(self):
        oo.pyoorb.oorb_init(ephemeris_fname=self.ephfile)

    def set_orbits(self, orbit_obj):
        """Set the orbits, to be used to generate ephemerides.

        Immediately calls self._convertOorbElem to translate to the 'packed' oorb format.

        Parameters
        ----------
        orbit_obj : `rubin_sim.moving_objects.Orbits`
           The orbits to use to generate ephemerides.
        """
        if len(orbit_obj) == 0:
            raise ValueError("There are no orbits in the Orbit object.")
        self._convert_to_oorb_elem(orbit_obj.orbits, orbit_obj.orb_format)

    def _convert_to_oorb_elem(self, orbit_dataframe, orb_format):
        """Convert orbital elements into the numpy fortran-format array OpenOrb requires.

        The OpenOrb element format is a single array with elemenets:
        0 : orbitId (cannot be a string)
        1-6 : orbital elements, using radians for angles
        7 : element 'type' code (1 = CAR, 2 = COM, 3 = KEP, 4 = DELauny, 5 = EQX (equinoctial))
        8 : epoch
        9 : timescale for epoch (1 = UTC, 2 = UT1, 3 = TT, 4 = TAI : always assumes TT)
        10 : magHv
        11 : g

        Sets self.oorb_elem, the orbit parameters in an array formatted for OpenOrb.
        """
        oorb_elem = np.zeros([len(orbit_dataframe), 12], dtype=np.double, order="F")
        # Put in simple values for objid, or add method to test if any obj_id is a string.
        # NOTE THAT THIS MEANS WE'VE LOST THE OBJID
        oorb_elem[:, 0] = np.arange(0, len(orbit_dataframe), dtype=int) + 1
        # Add the appropriate element and epoch types:
        oorb_elem[:, 7] = np.zeros(len(orbit_dataframe), float) + self.elem_type[orb_format]
        oorb_elem[:, 9] = np.zeros(len(orbit_dataframe), float) + self.time_scales["TT"]
        # Convert other elements INCLUDING converting inclination, node, argperi to RADIANS
        if orb_format == "KEP":
            oorb_elem[:, 1] = orbit_dataframe["a"]
            oorb_elem[:, 2] = orbit_dataframe["e"]
            oorb_elem[:, 3] = np.radians(orbit_dataframe["inc"])
            oorb_elem[:, 4] = np.radians(orbit_dataframe["Omega"])
            oorb_elem[:, 5] = np.radians(orbit_dataframe["argPeri"])
            oorb_elem[:, 6] = np.radians(orbit_dataframe["meanAnomaly"])
        elif orb_format == "COM":
            oorb_elem[:, 1] = orbit_dataframe["q"]
            oorb_elem[:, 2] = orbit_dataframe["e"]
            oorb_elem[:, 3] = np.radians(orbit_dataframe["inc"])
            oorb_elem[:, 4] = np.radians(orbit_dataframe["Omega"])
            oorb_elem[:, 5] = np.radians(orbit_dataframe["argPeri"])
            oorb_elem[:, 6] = orbit_dataframe["tPeri"]
        elif orb_format == "CAR":
            oorb_elem[:, 1] = orbit_dataframe["x"]
            oorb_elem[:, 2] = orbit_dataframe["y"]
            oorb_elem[:, 3] = orbit_dataframe["z"]
            oorb_elem[:, 4] = orbit_dataframe["xdot"]
            oorb_elem[:, 5] = orbit_dataframe["ydot"]
            oorb_elem[:, 6] = orbit_dataframe["zdot"]
        else:
            raise ValueError("Unknown orbit format %s: should be COM, KEP or CAR." % orb_format)
        oorb_elem[:, 8] = orbit_dataframe["epoch"]
        oorb_elem[:, 10] = orbit_dataframe["H"]
        oorb_elem[:, 11] = orbit_dataframe["g"]
        self.oorb_elem = oorb_elem
        self.orb_format = orb_format

    def convert_from_oorb_elem(self):
        """Translate pyoorb-style orbital element array back into dataframe.

        Parameters
        ----------
        oorbElem : `np.ndarray`
            The orbital elements in OpenOrb format.

        Returns
        -------
        new_orbits : `pd.DataFrame`
            A DataFrame with the appropriate subset of columns relating to orbital elements.
        """
        if self.orb_format == "KEP":
            new_orbits = pd.DataFrame(
                self.oorb_elem.copy(),
                columns=[
                    "oorbId",
                    "a",
                    "e",
                    "inc",
                    "Omega",
                    "argPeri",
                    "meanAnomaly",
                    "elem_type",
                    "epoch",
                    "epoch_type",
                    "H",
                    "g",
                ],
            )
            new_orbits["meanAnomaly"] = np.degrees(new_orbits["meanAnomaly"])
        elif self.orb_format == "COM":
            new_orbits = pd.DataFrame(
                self.oorb_elem.copy(),
                columns=[
                    "oorbId",
                    "q",
                    "e",
                    "inc",
                    "Omega",
                    "argPeri",
                    "tPeri",
                    "elem_type",
                    "epoch",
                    "epoch_type",
                    "H",
                    "g",
                ],
            )
        elif self.orb_format == "CAR":
            new_orbits = pd.DataFrame(
                self.oorb_elem.copy(),
                columns=[
                    "oorbId",
                    "x",
                    "y",
                    "z",
                    "xdot",
                    "ydot",
                    "zdot",
                    "elem_type",
                    "epoch",
                    "epoch_type",
                    "H",
                    "g",
                ],
            )
        else:
            raise ValueError("Unknown orbit format %s: should be COM, KEP or CAR." % self.orb_format)
        # Convert from radians to degrees.
        if self.orb_format == "KEP" or self.orb_format == "COM":
            new_orbits["inc"] = np.degrees(new_orbits["inc"])
            new_orbits["Omega"] = np.degrees(new_orbits["Omega"])
            new_orbits["argPeri"] = np.degrees(new_orbits["argPeri"])
        # Drop columns we don't need and don't include in our standard columns.
        del new_orbits["elem_type"]
        del new_orbits["epoch_type"]
        del new_orbits["oorbId"]
        # To incorporate with original Orbits object, need to swap back to original obj_ids
        # as well as put back in original SEDs.
        return new_orbits

    def convert_orbit_format(self, orb_format="CAR"):
        """Convert orbital elements from the format in orbitObj into 'format'.

        Parameters
        ----------
        format : `str`, optional
            Format to convert orbital elements into.
        """
        oorb_elem, err = oo.pyoorb.oorb_element_transformation(
            in_orbits=self.oorb_elem, in_element_type=self.elem_type[orb_format]
        )
        if err != 0:
            raise RuntimeError("Oorb returned error %s" % (err))
        del self.oorb_elem
        self.oorb_elem = oorb_elem
        self.orb_format = orb_format
        return

    def _convert_times(self, times, time_scale="UTC"):
        """Generate an oorb-format array of the times desired for the ephemeris generation.

        Parameters
        ----------
        times : `np.ndarray` or `float`
            The ephemeris times (MJD) desired
        time_scale : `str`, optional
            The timescale (UTC, UT1, TT, TAI) of the ephemeris MJD values. Default = UTC, MJD.

        Returns
        -------
        eph_times : `np.ndarray`
            The oorb-formatted 'eph_times' array.
        """
        if isinstance(times, float):
            times = np.array([times])
        if len(times) == 0:
            raise ValueError("Got zero times to convert for OpenOrb")
        eph_times = np.array(
            list(zip(times, repeat(self.time_scales[time_scale], len(times)))),
            dtype="double",
            order="F",
        )
        return eph_times

    def _generate_oorb_ephs_full(self, eph_times, obscode="I11", eph_mode="N"):
        """Generate full set of ephemeris output values using Oorb.

        Parameters
        ----------
        ephtimes : `np.ndarray`
            Ephemeris times in oorb format (see self.convertTimes)
        obscode : `int` or `str`, optional
            The observatory code for ephemeris generation. Default=I11 (Cerro Pachon).

        Returns
        -------
        ephemerides : `np.ndarray`
            The oorb-formatted ephemeris array.
        """
        oorb_ephems, err = oo.pyoorb.oorb_ephemeris_full(
            in_orbits=self.oorb_elem,
            in_obscode=obscode,
            in_date_ephems=eph_times,
            in_dynmodel=eph_mode,
        )
        if err != 0:
            raise RuntimeError("Oorb returned error %s" % (err))
        return oorb_ephems

    def _convert_oorb_ephs_full(self, oorb_ephs, by_object=True):
        """Converts oorb ephemeris array to numpy recarray, with labeled columns.

        The oorb ephemeris array is a 3-d array organized as: (object / times / eph@time)
        [objid][time][ephemeris information @ that time] with ephemeris elements
        ! (1) modified julian date
        ! (2) right ascension (deg)
        ! (3) declination (deg)
        ! (4) dra/dt sky-motion (deg/day, including cos(dec) factor)
        ! (5) ddec/dt sky-motion (deg/day)
        ! (6) solar phase angle (deg)
        ! (7) solar elongation angle (deg)
        ! (8) heliocentric distance (au)
        ! (9) geocentric distance (au)
        ! (10) predicted apparent V-band magnitude
        ! (11) position angle for direction of motion (deg)
        ! (12) topocentric ecliptic longitude (deg)
        ! (13) topocentric ecliptic latitude (deg)
        ! (14) opposition-centered topocentric ecliptic longitude (deg)
        ! (15) opposition-centered topocentric ecliptic latitude (deg)
        ! (16) heliocentric ecliptic longitude (deg)
        ! (17) heliocentric ecliptic latitude (deg)
        ! (18) opposition-centered heliocentric ecliptic longitude (deg)
        ! (19) opposition-centered heliocentric ecliptic latitude (deg)
        ! (20) topocentric object altitude (deg)
        ! (21) topocentric solar altitude (deg)
        ! (22) topocentric lunar altitude (deg)
        ! (23) lunar phase [0...1]
        ! (24) lunar elongation (deg, distance between the target and the Moon)
        ! (25) heliocentric ecliptic cartesian x coordinate for the object (au)
        ! (26) heliocentric ecliptic cartesian y coordinate for the object (au)
        ! (27) heliocentric ecliptic cartesian z coordinate for the objects (au)
        ! (28) heliocentric ecliptic cartesian x rate for the object (au/day))
        ! (29) heliocentric ecliptic cartesian y rate for the object (au/day)
        ! (30) heliocentric ecliptic cartesian z rate for the objects (au/day)
        ! (31) heliocentric ecliptic cartesian coordinates for the observatory (au)
        ! (32) heliocentric ecliptic cartesian coordinates for the observatory (au)
        ! (33) heliocentric ecliptic cartesian coordinates for the observatory (au)
        ! (34) true anomaly (currently only a dummy value)

        Here we convert to a numpy recarray, grouped either by object (default)
        or by time (if by_object=False).
        The resulting numpy recarray is composed of columns (of each ephemeris element),
        where each column is 2-d array with first axes either 'object' or 'time'.
        - if by_object = True : [ephemeris elements][object][time]
        (i.e. the 'ra' column = 2-d array, where the [0] axis (length) equals the number of ephTimes)
        - if by_object = False : [ephemeris elements][time][object]
        (i.e. the 'ra' column = 2-d arrays, where the [0] axis (length) equals the number of objects)

        Parameters
        ----------
        oorb_ephs : `np.ndarray`
            The oorb-formatted ephemeris values
        by_object : `bool`, optional
            If True (default), resulting converted ephemerides are grouped by object.
            If False, resulting converted ephemerides are grouped by time.

        Returns
        -------
        ephemerides : `np.ndarray`
            The re-arranged ephemeris values, in a 3-d array.
        """
        ephs = np.swapaxes(oorb_ephs, 2, 0)
        velocity = np.sqrt(ephs[3] ** 2 + ephs[4] ** 2)
        if by_object:
            ephs = np.swapaxes(ephs, 2, 1)
            velocity = np.swapaxes(velocity, 1, 0)
        # Create a numpy recarray.
        names = [
            "time",
            "ra",
            "dec",
            "dradt",
            "ddecdt",
            "phase",
            "solarelon",
            "helio_dist",
            "geo_dist",
            "magV",
            "pa",
            "topo_lon",
            "topo_lat",
            "opp_topo_lon",
            "opp_topo_lat",
            "helio_lon",
            "helio_lat",
            "opp_helio_lon",
            "opp_helio_lat",
            "topo_obj_alt",
            "topo_solar_alt",
            "topo_lunar_alt",
            "lunar_phase",
            "lunar_dist",
            "helio_x",
            "helio_y",
            "helio_z",
            "helio_dx",
            "helio_dy",
            "helio_dz",
            "obs_helio_x",
            "obs_helio_y",
            "obs_helio_z",
            "trueAnom",
        ]
        arraylist = []
        for i, n in enumerate(names):
            arraylist.append(ephs[i])
        arraylist.append(velocity)
        names.append("velocity")
        ephs = np.rec.fromarrays(arraylist, names=names)
        return ephs

    def _generate_oorb_ephs_basic(self, eph_times, obscode="I11", eph_mode="N"):
        """Generate ephemerides using OOrb with two body mode.

        Parameters
        ----------
        ephtimes : `np.ndarray`
            Ephemeris times in oorb format (see self.convertTimes).
        obscode : `int` or `str`, optional
            The observatory code for ephemeris generation. Default=I11 (Cerro Pachon).

        Returns
        -------
        oorb_ephems : `np.ndarray`
            The oorb-formatted ephemeris array.
        """
        oorb_ephems, err = oo.pyoorb.oorb_ephemeris_basic(
            in_orbits=self.oorb_elem,
            in_obscode=obscode,
            in_date_ephems=eph_times,
            in_dynmodel=eph_mode,
        )
        if err != 0:
            raise RuntimeError("Oorb returned error %s" % (err))
        return oorb_ephems

    def _convert_oorb_ephs_basic(self, oorb_ephs, by_object=True):
        """Converts oorb ephemeris array to numpy recarray, with labeled columns.

        The oorb ephemeris array is a 3-d array organized as: (object / times / eph@time)
        [objid][time][ephemeris information @ that time] with ephemeris elements
        ! (1) modified julian date
        ! (2) right ascension (deg)
        ! (3) declination (deg)
        ! (4) dra/dt sky-motion (deg/day, including cos(dec) factor)
        ! (5) ddec/dt sky-motion (deg/day)
        ! (6) solar phase angle (deg)
        ! (7) solar elongation angle (deg)
        ! (8) heliocentric distance (au)
        ! (9) geocentric distance (au)
        ! (10) predicted apparent V-band magnitude
        ! (11) true anomaly (currently only a dummy value)

        Here we convert to a numpy recarray, grouped either by object (default)
        or by time (if by_object=False).
        The resulting numpy recarray is composed of columns (of each ephemeris element),
        where each column is 2-d array with first axes either 'object' or 'time'.
        - if by_object = True : [ephemeris elements][object][time]
        (i.e. the 'ra' column = 2-d array, where the [0] axis (length) equals the number of ephTimes)
        - if by_object = False : [ephemeris elements][time][object]
        (i.e. the 'ra' column = 2-d arrays, where the [0] axis (length) equals the number of objects)

        Parameters
        ----------
        oorb_ephs : `np.ndarray`
            The oorb-formatted ephemeris values
        by_object : `bool`, optional
            If True (default), resulting converted ephemerides are grouped by object.
            If False, resulting converted ephemerides are grouped by time.

        Returns
        -------
        ephs : `np.ndarray`
            The re-arranged ephemeris values, in a 3-d array.
        """
        ephs = np.swapaxes(oorb_ephs, 2, 0)
        velocity = np.sqrt(ephs[3] ** 2 + ephs[4] ** 2)
        if by_object:
            ephs = np.swapaxes(ephs, 2, 1)
            velocity = np.swapaxes(velocity, 1, 0)
        # Create a numpy recarray.
        names = [
            "time",
            "ra",
            "dec",
            "dradt",
            "ddecdt",
            "phase",
            "solarelon",
            "helio_dist",
            "geo_dist",
            "magV",
            "trueAnomaly",
        ]
        arraylist = []
        for i, n in enumerate(names):
            arraylist.append(ephs[i])
        arraylist.append(velocity)
        names.append("velocity")
        ephs = np.rec.fromarrays(arraylist, names=names)
        return ephs

    def generate_ephemerides(
        self,
        times,
        time_scale="UTC",
        obscode="I11",
        by_object=True,
        eph_mode="nbody",
        eph_type="basic",
    ):
        """Calculate ephemerides for all orbits at times `times`.

        This is a public method, wrapping self._convert_times, self._generateOorbEphs
        and self._convertOorbEphs (which include dealing with oorb-formatting of arrays).

        The return ephemerides are in a numpy recarray, with axes
        - if by_object = True : [ephemeris values][object][@time]
        (i.e. the 'ra' column = 2-d array, where the [0] axis (length) equals the number of eph_times)
        - if by_object = False : [ephemeris values][time][@object]
        (i.e. the 'ra' column = 2-d arrays, where the [0] axis (length) equals the number of objects)

        The ephemeris values returned to the user (== columns of the recarray) are:
        ['delta', 'ra', 'dec', 'magV', 'time', 'dradt', 'ddecdt', 'phase', 'solarelon', 'velocity']
        where positions/angles are all in degrees, velocities are deg/day, and delta is the
        distance between the Earth and the object in AU.

        Parameters
        ----------
        ephtimes : `np.ndarray`
            Ephemeris times in oorb format (see self.convertTimes)
        obscode : `int` or `str`, optional
            The observatory code for ephemeris generation. Default=807 (Cerro Tololo).
        by_object : `bool`, optional
            If True (default), resulting converted ephemerides are grouped by object.
            If False, resulting converted ephemerides are grouped by time.
        eph_mode : `str`, optional
            Dynamical model to use for ephemeris generation - nbody or 2body.
            Accepts 'nbody', '2body', 'N' or '2'. Default nbody.
        eph_type : `str`, optional
            Generate full (more data) ephemerides or basic (less data) ephemerides.
            Default basic.

        Returns
        -------
        ephemerides : `np.ndarray`
            The ephemeris values, organized as chosen by the user.
        """
        if eph_mode.lower() in ("nbody", "n"):
            eph_mode = "N"
        elif eph_mode.lower() in ("2body", "2"):
            eph_mode = "2"
        else:
            raise ValueError("eph_mode should be 2body or nbody (or '2' or 'N').")

        # t = time.time()
        eph_times = self._convert_times(times, time_scale=time_scale)
        if eph_type.lower() == "basic":
            # oorb_ephs = self._generate_oorb_ephs_basic(eph_times, obscode=obscode, eph_mode=eph_mode)
            oorb_ephs, err = oo.pyoorb.oorb_ephemeris_basic(
                in_orbits=self.oorb_elem,
                in_obscode=obscode,
                in_date_ephems=eph_times,
                in_dynmodel=eph_mode,
            )
            ephs = self._convert_oorb_ephs_basic(oorb_ephs, by_object=by_object)
        elif eph_type.lower() == "full":
            oorb_ephs = self._generate_oorb_ephs_full(eph_times, obscode=obscode, eph_mode=eph_mode)
            ephs = self._convert_oorb_ephs_full(oorb_ephs, by_object=by_object)
        else:
            raise ValueError("eph_type must be full or basic")
        # dt, t = dtime(t)
        # logging.debug("# Calculating ephemerides for %d objects over %d times required %f seconds"
        #              % (len(self.oorb_elem), len(times), dt))
        return ephs

    def propagate_orbits(self, new_epoch, eph_mode="nbody"):
        """Propagate orbits from self.orbits.epoch to new epoch (MJD TT).

        Parameters
        ----------
        new_epoch : `float`
            MJD TT time for new epoch.
        """
        new_epoch = self._convert_times(new_epoch, time_scale="TT")
        if eph_mode.lower() in ("nbody", "n"):
            eph_mode = "N"
        elif eph_mode.lower() in ("2body", "2"):
            eph_mode = "2"
        else:
            raise ValueError("eph_mode should be 2body or nbody (or '2' or 'N').")

        new_oorb_elem, err = oo.pyoorb.oorb_propagation(
            in_orbits=self.oorb_elem, in_dynmodel=eph_mode, in_epoch=new_epoch
        )
        if err != 0:
            raise RuntimeError("Orbit propagation returned error %d" % err)
        self.oorb_elem = new_oorb_elem
        return



__all__ = ("StarCountMassMetric",)

import numpy as np

from rubin_sim.maf.metrics import BaseMetric

from .star_counts import *

# Example for CountMassMetric
# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Motivation: The distances to stars in LSST will be signficant enough that the structure of the Galaxy will be readily apparent because of its influence on the number of stars in a given field. Any metric concerned with the number of potential objects to be detected will need to feature not only the effects of the cadence but also the number of objects per field.
# This metric identifies the number of stars in a given field in a particular mass range that will be fainter than the saturation limit of 16th magnitude and still bright enough to have noise less than 0.03 mag. M1 and M2 are the low and high limits of the mass range in solar masses. 'band' is the band for the observations to be made in.
# Requires StarCounts.StarCounts


class StarCountMassMetric(BaseMetric):
    """Find the number of stars in a given field in the mass range fainter than magnitude 16 and bright enough to have noise less than 0.03 in a given band. M1 and m2 are the upper and lower limits of the mass range. 'band' is the band to be observed."""

    def __init__(self, **kwargs):
        self.m1 = kwargs.pop("M1", 0.9)
        self.m2 = kwargs.pop("M2", 1.0)
        self.band = kwargs.pop("band", "i")
        super(StarCountMassMetric, self).__init__(col=[], **kwargs)

    def run(self, data_slice, slice_point=None):
        self.dec_col = np.degrees(data_slice[0][3])
        self.ra_col = np.degrees(data_slice[0][2])
        return starcount_bymass.starcount_bymass(self.ra_col, self.dec_col, self.m1, self.m2, self.band)



# Base class for metrics - defines methods which must be implemented.
# If a metric calculates a vector or list at each gridpoint, then there
#  should be additional 'reduce_*' functions defined, to convert the vector
#  into scalar (and thus plottable) values at each gridpoint.
# The philosophy behind keeping the vector instead of the scalar at each gridpoint
#  is that these vectors may be expensive to compute; by keeping/writing the full
#  vector we permit multiple 'reduce' functions to be executed on the same data.

__all__ = ("MetricRegistry", "BaseMetric", "ColRegistry")

import inspect
import warnings

import numpy as np

from rubin_sim.maf.stackers.get_col_info import ColInfo


class MetricRegistry(type):
    """
    Meta class for metrics, to build a registry of metric classes.
    """

    def __init__(cls, name, bases, dict):
        super(MetricRegistry, cls).__init__(name, bases, dict)
        if not hasattr(cls, "registry"):
            cls.registry = {}
        modname = inspect.getmodule(cls).__name__
        if modname.startswith("rubin_sim.maf.metrics"):
            modname = ""
        else:
            if len(modname.split(".")) > 1:
                modname = ".".join(modname.split(".")[:-1]) + "."
            else:
                modname = modname + "."
        metricname = modname + name
        if metricname in cls.registry:
            warnings.warn("Redefining metric %s! (there are >1 metrics with the same name)" % (metricname))
        if metricname not in ["BaseMetric", "SimpleScalarMetric"]:
            cls.registry[metricname] = cls

    def get_class(cls, metricname):
        return cls.registry[metricname]

    def help(cls, doc=False):
        for metricname in sorted(cls.registry):
            if not doc:
                print(metricname)
            if doc:
                print("---- ", metricname, " ----")
                print(inspect.getdoc(cls.registry[metricname]))

    def help_metric(cls, metricname):
        print(metricname)
        print(inspect.getdoc(cls.registry[metricname]))
        k = inspect.signature(cls.registry[metricname])
        print(" Metric __init__ keyword args and defaults: ")
        print(k)


class ColRegistry:
    """
    ColRegistry tracks the columns needed for all metric objects
    (kept internally in a set).

    ColRegistry.col_set : a set of all unique columns required for metrics.
    ColRegistry.dbCols : the subset of these which come from the database.
    ColRegistry.stackerCols : the dictionary of [columns: stacker class].
    """

    col_info = ColInfo()

    def __init__(self):
        self.col_set = set()
        self.db_set = set()
        self.stacker_dict = {}

    def clear_reg(self):
        """Clear the registry"""
        self.__init__()

    def add_cols(self, col_array):
        """Add the columns in ColArray into the ColRegistry.

        Add the columns in col_array into the ColRegistry set (self.col_set)
        and identifies their source, using ColInfo
        (rubin_sim.maf.stackers.getColInfo).

        Parameters
        ----------
        col_array : `list`
            list of columns used in a metric.
        """
        for col in col_array:
            if col is not None:
                self.col_set.add(col)
                source = self.col_info.get_data_source(col)
                if source == self.col_info.default_data_source:
                    self.db_set.add(col)
                else:
                    if col not in self.stacker_dict:
                        self.stacker_dict[col] = source


class BaseMetric(metaclass=MetricRegistry):
    """
    Base class for the metrics.
    Sets up some basic functionality for the MAF framework:
    after __init__ every metric will record the columns (and stackers)
    it requires into the column registry, and the metric_name,
    metric_dtype, and units for the metric will be set.

    Parameters
    ----------
    col : `str` or `list` [`str`]
        Names of the data columns that the metric will use.
        The columns required for each metric is tracked in the ColRegistry, and used to retrieve data
        from the opsim database. Can be a single string or a list.
    metric_name : `str`
        Name to use for the metric (optional - if not set, will be derived).
    maps : `list` [`rubin_sim.maf.maps`]
        The maps that the metric will need (passed from the slicer).
    units : `str`
        The units for the value returned by the metric (optional - if not set,
        will be derived from the ColInfo).
    metric_dtype : `str`
        The type of value returned by the metric - 'int', 'float', 'object'.
        If not set, will be derived by introspection.
    badval : `float`
        The value indicating "bad" values calculated by the metric.
    """

    col_registry = ColRegistry()
    col_info = ColInfo()

    def __init__(
        self,
        col=None,
        metric_name=None,
        maps=None,
        units=None,
        metric_dtype=None,
        badval=-666,
        mask_val=None,
    ):
        # Turn cols into numpy array so we know we can iterate over the columns.
        self.col_name_arr = np.array(col, copy=False, ndmin=1)
        # To support simple metrics operating on a single column, set self.colname
        if len(self.col_name_arr) == 1:
            self.colname = self.col_name_arr[0]
        # Add the columns to the colRegistry.
        self.col_registry.add_cols(self.col_name_arr)
        # Set the maps that are needed:
        if maps is None:
            maps = []
        self.maps = maps
        # Value to return if the metric can't be computed
        self.badval = badval
        if mask_val is not None:
            self.mask_val = mask_val
        # Save a unique name for the metric.
        self.name = metric_name
        if self.name is None:
            # If none provided, construct our own from the class name and the data columns.
            self.name = (
                self.__class__.__name__.replace("Metric", "", 1)
                + " "
                + ", ".join(map(str, self.col_name_arr))
            )
        # Set up dictionary of reduce functions (may be empty).
        self.reduce_funcs = {}
        self.reduce_order = {}
        for i, r in enumerate(inspect.getmembers(self, predicate=inspect.ismethod)):
            if r[0].startswith("reduce"):
                # Remove the "reduce_" part of the reduce functions names.
                reducename = r[0].replace("reduce_", "", 1)
                self.reduce_funcs[reducename] = r[1]
                self.reduce_order[reducename] = i
        # Identify type of metric return value.
        if metric_dtype is not None:
            self.metric_dtype = metric_dtype
        elif len(list(self.reduce_funcs.keys())) > 0:
            self.metric_dtype = "object"
        else:
            self.metric_dtype = "float"
        # Set physical units, for plotting purposes.
        if units is None:
            units = " ".join([self.col_info.get_units(col_name) for col_name in self.col_name_arr])
            if len(units.replace(" ", "")) == 0:
                units = ""
        self.units = units
        # Add the ability to set a comment
        # (that could be propagated automatically to a benchmark's
        # display caption).
        self.comment = None

        # Default to only return one metric value per slice
        self.shape = 1

    def run(self, data_slice, slice_point=None):
        """Calculate metric values.

        Parameters
        ----------
        data_slice : `numpy.ndarray`, (N,)
           Values passed to metric by the slicer, which the metric will
           use to calculate metric values at each slice_point.
        slice_point : `dict` or None
           Dictionary of slice_point metadata passed to each metric.
           E.g. the ra/dec of the healpix pixel or opsim fieldId.

        Returns
        -------
        metricValue: `int` `float` or `object`
            The metric value at each slice_point.
        """
        raise NotImplementedError("Please implement your metric calculation.")



__all__ = ("TrilegalDensityMap",)

import os

import healpy as hp
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import _build_tree, _hpid2_ra_dec, _xyz_from_ra_dec

from . import BaseMap


class TrilegalDensityMap(BaseMap):
    """
    Return the cumulative stellar luminosity function for each slice_point. Units of stars per sq degree.

    Parameters
    ----------
    filtername : `str`, opt
        Filter to use. Options of u,g,r,i,z,y. Default r.
    nside : `int`, opt
        The HEALpix nside (can be 64 or 128). Default 64.
    ext : `bool`, opt
        Use the full sky maps. Default True.
    """

    def __init__(self, filtername="r", nside=64, ext=True):
        self.map_dir = os.path.join(get_data_dir(), "maps", "TriMaps")
        self.filtername = filtername
        self.keynames = [
            f"starLumFunc_{self.filtername}",
            f"starMapBins_{self.filtername}",
        ]
        self.nside = nside
        self.ext = ext

    def _read_map(self):
        if self.ext:
            filename = "TRIstarDensity_%s_nside_%i_ext.npz" % (
                self.filtername,
                self.nside,
            )
        else:
            filename = "TRIstarDensity_%s_nside_%i.npz" % (self.filtername, self.nside)
        star_map = np.load(os.path.join(self.map_dir, filename))
        self.star_map = star_map["starDensity"].copy()
        self.star_map_bins = star_map["bins"].copy()
        self.starmap_nside = hp.npix2nside(np.size(self.star_map[:, 0]))
        # note, the trilegal maps are in galactic coordinates, and nested healpix.
        gal_l, gal_b = _hpid2_ra_dec(self.nside, np.arange(hp.nside2npix(self.nside)), nest=True)

        # Convert that to RA,dec. Then do nearest neighbor lookup.
        c = SkyCoord(l=gal_l * u.rad, b=gal_b * u.rad, frame="galactic").transform_to("icrs")
        ra = c.ra.rad
        dec = c.dec.rad

        self.tree = _build_tree(ra, dec)

    def run(self, slice_points):
        self._read_map()

        x, y, z = _xyz_from_ra_dec(slice_points["ra"], slice_points["dec"])

        dist, indices = self.tree.query(list(zip(x, y, z)))

        slice_points["starLumFunc_%s" % self.filtername] = self.star_map[indices, :]
        slice_points["starMapBins_%s" % self.filtername] = self.star_map_bins
        return slice_points



__all__ = ("VisitGroupsMetric", "PairFractionMetric")


import numpy as np

from .base_metric import BaseMetric

# Example of more complex metric
# Takes multiple columns of data (although 'night' could be calculable from 'expmjd')
# Returns variable length array of data
# Uses multiple reduce functions


class PairFractionMetric(BaseMetric):
    """What fraction of observations are part of a pair.

    Note, an observation can be a member of more than one "pair". For example,
    t=[0, 5, 30], all observations would be considered part of a pair because they
    all have an observation within the given window to pair with (the observation at t=30
    pairs twice).

    Parameters
    ----------
    min_gap : float, optional
        Minimum time to consider something part of a pair (minutes). Default 15.
    max_gap : float, optional
        Maximum time to consider something part of a pair (minutes). Default 90.
    """

    def __init__(
        self, mjd_col="observationStartMJD", metric_name="PairFraction", min_gap=15.0, max_gap=90.0, **kwargs
    ):
        self.mjd_col = mjd_col
        self.min_gap = min_gap / 60.0 / 24.0
        self.max_gap = max_gap / 60.0 / 24.0
        units = ""
        super(PairFractionMetric, self).__init__(
            col=[mjd_col], metric_name=metric_name, units=units, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        nobs = np.size(data_slice[self.mjd_col])
        times = np.sort(data_slice[self.mjd_col])

        # Check which ones have a forard match
        t_plus = times + self.max_gap
        t_minus = times + self.min_gap
        ind1 = np.searchsorted(times, t_plus)
        ind2 = np.searchsorted(times, t_minus)
        # If ind1 and ind2 are the same, there is no pairable image for that exposure
        diff1 = ind1 - ind2

        # Check which have a back match
        t_plus = times - self.max_gap
        t_minus = times - self.min_gap
        ind1 = np.searchsorted(times, t_plus)
        ind2 = np.searchsorted(times, t_minus)

        diff2 = ind1 - ind2

        # The exposure has a pair ahead or behind
        is_paired = np.where((diff1 != 0) | (diff2 != 0))[0]
        result = np.size(is_paired) / float(nobs)
        return result


class VisitGroupsMetric(BaseMetric):
    """Count the number of visits per night within delta_t_min and delta_t_max."""

    def __init__(
        self,
        time_col="observationStartMJD",
        nights_col="night",
        metric_name="VisitGroups",
        delta_t_min=15.0 / 60.0 / 24.0,
        delta_t_max=90.0 / 60.0 / 24.0,
        min_n_visits=2,
        window=30,
        min_n_nights=3,
        **kwargs,
    ):
        """
        Instantiate metric.

        'time_col' = column with the time of the visit (default expmjd),
        'nights_col' = column with the night of the visit (default night),
        'delta_t_min' = minimum time of window: units are days (default 15 min),
        'delta_t_max' = maximum time of window: units are days (default 90 min),
        'min_n_visits' = the minimum number of visits within a night (with spacing between delta_t_min/max
        from any other visit) required,
        'window' = the number of nights to consider within a window (for reduce methods),
        'min_n_nights' = the minimum required number of nights within window to make a full 'group'.
        """
        self.times = time_col
        self.nights = nights_col
        eps = 1e-10
        self.delta_tmin = float(delta_t_min) - eps
        self.delta_tmax = float(delta_t_max)
        self.min_n_visits = int(min_n_visits)
        self.window = int(window)
        self.min_n_nights = int(min_n_nights)
        super(VisitGroupsMetric, self).__init__(
            col=[self.times, self.nights], metric_name=metric_name, **kwargs
        )
        self.reduce_order = {
            "Median": 0,
            "NNightsWithNVisits": 1,
            "NVisitsInWindow": 2,
            "NNightsInWindow": 3,
            "NLunations": 4,
            "MaxSeqLunations": 5,
        }
        self.comment = "Evaluation of the number of visits within a night, with separations between "
        self.comment += "tMin %.1f and tMax %.1f minutes." % (
            self.delta_tmin * 24.0 * 60.0,
            self.delta_tmax * 24.0 * 60.0,
        )
        self.comment += "Groups of visits use a minimum number of visits per night of %d, " % (
            self.min_n_visits
        )
        self.comment += "and minimum number of nights of %d." % (self.min_n_nights)
        self.comment += "Two visits within this interval would count as 2. "
        self.comment += "Visits closer than tMin, paired with visits that do fall within tMin/tMax, "
        self.comment += "count as half visits. <br>"
        self.comment += "VisitsGroups_Median calculates the median number of visits between tMin/tMax for "
        self.comment += "all nights. <br>"
        self.comment += "VisitGroups_NNightsWithNNights calculates the number of nights that have at "
        self.comment += "least %d visits. <br>" % (self.min_n_visits)
        self.comment += "VisitGroups_NVisitsInWindow calculates the max number of visits within a window of "
        self.comment += "%d days. <br>" % (self.window)
        self.comment += "VisitGroups_NNightsInWindow calculates the max number of nights that have more "
        self.comment += "than %d visits within %d days. <br>" % (
            self.min_n_visits,
            self.window,
        )
        self.comment += "VisitGroups_NLunations calculates the number of lunations (30 days) that have "
        self.comment += "at least one group of more than %d nights with more than %d visits, within " % (
            self.min_n_nights,
            self.min_n_visits,
        )
        self.comment += "%d days. <br>" % (self.window)
        self.comment += "VisitGroups_MaxSeqLunations calculates the maximum sequential lunations that have "
        self.comment += 'at least one "group". <br>'

    def run(self, data_slice, slice_point=None):
        """
        Return a dictionary of:
        the number of visits within a night (within delta tmin/tmax of another visit),
        and the nights with visits > minNVisits.
        Count two visits which are within tmin of each other, but which have another visit
        within tmin/tmax interval, as one and a half (instead of two).

        So for example: 4 visits, where 1, 2, 3 were all within deltaTMax of each other, and 4 was later but
        within deltaTmax of visit 3 -- would give you 4 visits. If visit 1 and 2 were closer together
        than deltaTmin, the two would be counted as 1.5 visits together (if only 1 and 2 existed,
        then there would be 0 visits as none would be within the qualifying time interval).
        """
        uniquenights = np.unique(data_slice[self.nights])
        nights = []
        visit_num = []
        # Find the nights with visits within deltaTmin/max of one another and count the number of visits
        for n in uniquenights:
            condition = data_slice[self.nights] == n
            times = np.sort(data_slice[self.times][condition])
            nvisits = 0
            ntooclose = 0
            # Calculate difference between each visit and time of previous visit (tnext- tnow)
            timediff = np.diff(times)
            timegood = np.where(
                (timediff <= self.delta_tmax) & (timediff >= self.delta_tmin),
                True,
                False,
            )
            timetooclose = np.where(timediff < self.delta_tmin, True, False)
            if len(timegood) > 1:
                # Count visits for all but last index in timediff
                for tg1, ttc1, tg2, ttc2 in zip(
                    timegood[:-1], timetooclose[:-1], timegood[1:], timetooclose[1:]
                ):
                    if tg1:
                        nvisits += 1
                        if not tg2:
                            nvisits += 1
                    if ttc1:
                        ntooclose += 1
                        if not tg2 and not ttc2:
                            ntooclose += 1
                # Take care of last timediff
                if timegood[-1]:
                    nvisits += 2  # close out visit sequence
                if timetooclose[-1]:
                    ntooclose += 1
                    if not timegood[-2] and not timetooclose[-2]:
                        ntooclose += 1
            else:
                if timegood.size > 0:
                    nvisits += 2
            # Count up all visits for night.
            if nvisits > 0:
                nvisits = nvisits + ntooclose / 2.0
                visit_num.append(nvisits)
                nights.append(n)
        # Convert to numpy arrays.
        visit_num = np.array(visit_num)
        nights = np.array(nights)
        metricval = {"visits": visit_num, "nights": nights}
        if len(visit_num) == 0:
            return self.badval
        return metricval

    def reduce_median(self, metricval):
        """Reduce to median number of visits per night."""
        return np.median(metricval["visits"])

    def reduce_n_nights_with_n_visits(self, metricval):
        """Reduce to total number of nights with more than 'minNVisits' visits."""
        condition = metricval["visits"] >= self.min_n_visits
        return len(metricval["visits"][condition])

    def _in_window(self, visits, nights, night, window, min_n_visits):
        condition = (nights >= night) & (nights < night + window)
        condition2 = visits[condition] >= min_n_visits
        return visits[condition][condition2], nights[condition][condition2]

    def reduce_n_visits_in_window(self, metricval):
        """Reduce to max number of total visits on all nights with more than minNVisits,
        within any 'window' (default=30 nights)."""
        maxnvisits = 0
        for n in metricval["nights"]:
            vw, nw = self._in_window(
                metricval["visits"],
                metricval["nights"],
                n,
                self.window,
                self.min_n_visits,
            )
            maxnvisits = max((vw.sum(), maxnvisits))
        return maxnvisits

    def reduce_n_nights_in_window(self, metricval):
        """Reduce to max number of nights with more than minNVisits, within 'window' over all windows."""
        maxnights = 0
        for n in metricval["nights"]:
            vw, nw = self._in_window(
                metricval["visits"],
                metricval["nights"],
                n,
                self.window,
                self.min_n_visits,
            )
            maxnights = max(len(nw), maxnights)
        return maxnights

    def _in_lunation(self, visits, nights, lunation_start, lunation_length):
        condition = (nights >= lunation_start) & (nights < lunation_start + lunation_length)
        return visits[condition], nights[condition]

    def reduce_n_lunations(self, metricval):
        """Reduce to number of lunations (unique 30 day windows) that contain at least one 'group':
        a set of more than minNVisits per night, with more than minNNights of visits within 'window' time period.
        """
        lunation_length = 30
        lunations = np.arange(
            metricval["nights"][0],
            metricval["nights"][-1] + lunation_length / 2.0,
            lunation_length,
        )
        n_lunations = 0
        for l in lunations:
            # Find visits within lunation.
            vl, nl = self._in_lunation(metricval["visits"], metricval["nights"], l, lunation_length)
            for n in nl:
                # Find visits which are in groups within the lunation.
                vw, nw = self._in_window(vl, nl, n, self.window, self.min_n_visits)
                if len(nw) >= self.min_n_nights:
                    n_lunations += 1
                    break
        return n_lunations

    def reduce_max_seq_lunations(self, metricval):
        """Count the max number of sequential lunations (unique 30 day windows) that contain at least one 'group':
        a set of more than minNVisits per night, with more than minNNights of visits within 'window' time period.
        """
        lunation_length = 30
        lunations = np.arange(
            metricval["nights"][0],
            metricval["nights"][-1] + lunation_length / 2.0,
            lunation_length,
        )
        max_sequence = 0
        cur_sequence = 0
        for l in lunations:
            # Find visits within lunation.
            vl, nl = self._in_lunation(metricval["visits"], metricval["nights"], l, lunation_length)
            # If no visits this lunation:
            if len(vl) == 0:
                max_sequence = max(max_sequence, cur_sequence)
                cur_sequence = 0
            # Else, look to see if groups can be made from the visits.
            for n in nl:
                # Find visits which are in groups within the lunation.
                vw, nw = self._in_window(vl, nl, n, self.window, self.min_n_visits)
                # If there was a group within this lunation:
                if len(nw) >= self.min_n_nights:
                    cur_sequence += 1
                    break
                # Otherwise we're not in a sequence (anymore, or still).
                else:
                    max_sequence = max(max_sequence, cur_sequence)
                    cur_sequence = 0
        # Pick up last sequence if were in a sequence at last lunation.
        max_sequence = max(max_sequence, cur_sequence)
        return max_sequence



__all__ = ("DustMap",)

import warnings

from rubin_sim.maf.maps import BaseMap

from .ebv_hp import eb_vhp


class DustMap(BaseMap):
    """
    Compute the E(B-V) for each point in a given spatial distribution of slicePoints.

    Primarily, this calls eb_vhp to read a healpix map of E(B-V) values over the sky, then
    assigns ebv values to each slice_point. If the slicer is a healpix slicer, this is trivial.

    Parameters
    ----------
    interp : `bool`, opt
        Interpolate the dust map at each slice_point (True) or just use the nearest value (False).
        Default is False.
    nside : `int`, opt
        Default nside value to read the dust map from disk. Primarily useful if the slicer is not
        a healpix slicer.
        Default 128.
    map_path : `str`, opt
        Define a path to the directory holding the dust map files.
        Default None, which uses RUBIN_SIM_DATA_DIR.
    """

    def __init__(self, interp=False, nside=128, map_path=None):
        """
        interp: should the dust map be interpolated (True) or just use the nearest value (False).
        """
        self.keynames = ["ebv"]
        self.interp = interp
        self.nside = nside
        self.map_path = map_path

    def run(self, slice_points):
        # If the slicer has nside, it's a healpix slicer so we can read the map directly
        if "nside" in slice_points:
            if slice_points["nside"] != self.nside:
                warnings.warn(
                    f"Slicer value of nside {slice_points['nside']} different "
                    f"from map value {self.nside}, using slicer value"
                )
            slice_points["ebv"] = eb_vhp(
                slice_points["nside"],
                pixels=slice_points["sid"],
                map_path=self.map_path,
            )
        # Not a healpix slicer, look up values based on RA,dec with possible interpolation
        else:
            slice_points["ebv"] = eb_vhp(
                self.nside,
                ra=slice_points["ra"],
                dec=slice_points["dec"],
                interp=self.interp,
                map_path=self.map_path,
            )

        return slice_points



import unittest

import numpy as np
import pandas as pd

import rubin_sim.maf.metrics as metrics


class TestMoMetrics1(unittest.TestCase):
    def setUp(self):
        # Set up some sso_obs data to test the metrics on.
        # Note that sso_obs is a numpy recarray.
        # The expected set of columns in sso_obs is:
        # cols = ['observationStartMJD', 'night', 'fieldRA', 'fieldDec',
        #        'rotSkyPos', 'filter',
        #        'visitExposureTime', 'seeingFwhmGeom', 'fiveSigmaDepth',
        #        'solarElong',
        #        'delta', 'ra', 'dec', 'magV', 'time', 'dradt', 'ddecdt',
        #        'phase', 'solarelon',
        #        'velocity', 'magFilter', 'dmagColor', 'dmagTrail',
        #        'dmagDetect']
        # And stackers will often add
        # addCols = ['appMag', 'magLimit', 'snr', 'vis']

        # Test metrics using sso_obs for a particular object.
        times = np.array(
            [0.1, 0.2, 0.3, 1.1, 1.3, 5.1, 7.1, 7.2, 7.3, 10.1, 10.2, 10.3, 13.1, 13.5],
            dtype="float",
        )
        sso_obs = np.recarray(
            [len(times)],
            dtype=(
                [
                    ("time", "<f8"),
                    ("ra", "<f8"),
                    ("dec", "<f8"),
                    ("appMag", "<f8"),
                    ("observationStartMJD", "<f8"),
                    ("night", "<f8"),
                    ("magLimit", "<f8"),
                    ("SNR", "<f8"),
                    ("vis", "<f8"),
                ]
            ),
        )

        sso_obs["time"] = times
        sso_obs["observationStartMJD"] = times
        sso_obs["night"] = np.floor(times)
        sso_obs["ra"] = np.arange(len(times))
        sso_obs["dec"] = np.arange(len(times))
        sso_obs["appMag"] = np.zeros(len(times), dtype="float") + 24.0
        sso_obs["magLimit"] = np.zeros(len(times), dtype="float") + 25.0
        sso_obs["SNR"] = np.zeros(len(times), dtype="float") + 5.0
        sso_obs["vis"] = np.zeros(len(times), dtype="float") + 1
        sso_obs["vis"][0:5] = 0
        self.sso_obs = sso_obs
        self.orb = None
        self.hval = 8.0

    def testn_obs_metric(self):
        n_obs_metric = metrics.NObsMetric(snr_limit=5)
        n_obs = n_obs_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(n_obs, len(self.sso_obs["time"]))
        n_obs_metric = metrics.NObsMetric(snr_limit=10)
        n_obs = n_obs_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(n_obs, 0)
        n_obs_metric = metrics.NObsMetric(snr_limit=None)
        n_obs = n_obs_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(n_obs, len(self.sso_obs["time"]) - 5)

    def testn_obs_no_singles_metric(self):
        n_obs_no_singles_metric = metrics.NObsNoSinglesMetric(snr_limit=5)
        n_obs = n_obs_no_singles_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(n_obs, len(self.sso_obs["time"]) - 1)

    def test_n_nights_metric(self):
        n_nights_metric = metrics.NNightsMetric(snr_limit=5)
        nnights = n_nights_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(nnights, len(np.unique(self.sso_obs["night"])))
        n_nights_metric = metrics.NNightsMetric(snr_limit=None)
        nnights = n_nights_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(nnights, len(np.unique(self.sso_obs["night"])) - 2)

    def test_arc_metric(self):
        arc_metric = metrics.ObsArcMetric(snr_limit=5)
        arc = arc_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(
            arc,
            self.sso_obs["observationStartMJD"][-1] - self.sso_obs["observationStartMJD"][0],
        )
        arc_metric = metrics.ObsArcMetric(snr_limit=None)
        arc = arc_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(
            arc,
            self.sso_obs["observationStartMJD"][-1] - self.sso_obs["observationStartMJD"][5],
        )

    def test_activity_over_period_metric(self):
        # cometary orbit format ok
        orb = np.recarray(
            1,
            dtype=(
                [
                    ("objId", (str, 20)),
                    ("q", float),
                    ("e", float),
                    ("inc", float),
                    ("Omega", float),
                    ("argPeri", float),
                    ("tPeri", float),
                    ("epoch", float),
                    ("H", float),
                    ("g", float),
                ]
            ),
        )

        orb["objId"] = "NESC00001HYj"
        orb["q"] = 1.00052
        orb["e"] = 0.028514
        orb["inc"] = 0.502477
        orb["Omega"] = 50.989131
        orb["argPeri"] = 55.091685
        orb["tPeri"] = 61046.627194 - 59850
        orb["epoch"] = 60973.799216 - 59850
        orb["H"] = 35.526041
        orb["g"] = 0.15
        o = pd.DataFrame(orb)
        activity_period_metric = metrics.ActivityOverPeriodMetric(bin_size=360, snr_limit=5)
        activity = activity_period_metric.run(self.sso_obs, o.iloc[0], self.hval)
        self.assertEqual(activity, 1.0)
        activity_period_metric = metrics.ActivityOverPeriodMetric(bin_size=720, snr_limit=5)
        activity = activity_period_metric.run(self.sso_obs, o.iloc[0], self.hval)
        self.assertEqual(activity, 1.0)
        activity_period_metric = metrics.ActivityOverPeriodMetric(bin_size=10, snr_limit=5)
        activity = activity_period_metric.run(self.sso_obs, o.iloc[0], self.hval)
        self.assertLess(activity, 0.03)
        # different type of orbit - currently should fail quietly
        orb = np.recarray(
            1,
            dtype=(
                [
                    ("objId", (str, 20)),
                    ("a", float),
                    ("e", float),
                    ("inc", float),
                    ("Omega", float),
                    ("argPeri", float),
                    ("meanAnomaly", float),
                    ("epoch", float),
                    ("H", float),
                    ("g", float),
                ]
            ),
        )
        orb["objId"] = "NESC00001HYj"
        orb["a"] = 1.029886
        orb["e"] = 0.028514
        orb["inc"] = 0.502477
        orb["Omega"] = 50.989131
        orb["argPeri"] = 55.091685
        orb["meanAnomaly"] = 291.321814
        orb["epoch"] = 60973.799216 - 59850
        orb["H"] = 35.526041
        orb["g"] = 0.15
        o = pd.DataFrame(orb)
        activity_period_metric = metrics.ActivityOverPeriodMetric(bin_size=360, snr_limit=5)
        activity = activity_period_metric.run(self.sso_obs, o.iloc[0], self.hval)
        self.assertEqual(activity, 1.0)
        activity_period_metric = metrics.ActivityOverPeriodMetric(bin_size=180, snr_limit=5)
        activity = activity_period_metric.run(self.sso_obs, o.iloc[0], self.hval)
        self.assertEqual(activity, 0.5)

    def tearDown(self):
        del self.sso_obs
        del self.orb
        del self.hval


class TestDiscoveryMetrics(unittest.TestCase):
    def setUp(self):
        rng = np.random.RandomState(61331)
        # Test metrics using sso_obs for a particular object.
        times = np.array(
            [0.1, 0.2, 0.9, 1.1, 1.3, 5.1, 7.1, 7.2, 7.5, 10.1, 10.2, 13.1, 13.5],
            dtype="float",
        )
        sso_obs = np.recarray(
            [len(times)],
            dtype=(
                [
                    ("time", "<f8"),
                    ("ra", "<f8"),
                    ("dec", "<f8"),
                    ("ec_lon", "<f8"),
                    ("ec_lat", "<f8"),
                    ("solar_elong", "<f8"),
                    ("appMag", "<f8"),
                    ("observationStartMJD", "<f8"),
                    ("night", "<f8"),
                    ("magLimit", "<f8"),
                    ("velocity", "<f8"),
                    ("SNR", "<f8"),
                    ("vis", "<f8"),
                    ("magFilter", "<f8"),
                    ("fiveSigmaDepth", "<f8"),
                    ("seeingFwhmGeom", "<f8"),
                    ("visitExposureTime", "<f8"),
                    ("dmagDetect", "<f8"),
                ]
            ),
        )

        sso_obs["time"] = times
        sso_obs["observationStartMJD"] = times
        sso_obs["night"] = np.floor(times)
        sso_obs["ra"] = np.arange(len(times))
        sso_obs["dec"] = np.arange(len(times)) + 5
        sso_obs["ec_lon"] = sso_obs["ra"] + 10
        sso_obs["ec_lat"] = sso_obs["dec"] + 20
        sso_obs["solar_elong"] = sso_obs["ra"] + 30
        sso_obs["appMag"] = np.zeros(len(times), dtype="float") + 24.0
        sso_obs["magFilter"] = np.zeros(len(times), dtype="float") + 24.0
        sso_obs["fiveSigmaDepth"] = np.zeros(len(times), dtype="float") + 25.0
        sso_obs["dmagDetect"] = np.zeros(len(times), dtype="float")
        sso_obs["magLimit"] = np.zeros(len(times), dtype="float") + 25.0
        sso_obs["SNR"] = np.zeros(len(times), dtype="float") + 5.0
        sso_obs["vis"] = np.zeros(len(times), dtype="float") + 1
        sso_obs["vis"][0:5] = 0
        sso_obs["velocity"] = rng.rand(len(times))
        sso_obs["seeingFwhmGeom"] = np.ones(len(times), "float")
        sso_obs["visitExposureTime"] = np.ones(len(times), "float") * 24.0
        self.sso_obs = sso_obs
        self.orb = np.recarray([len(times)], dtype=([("H", "<f8")]))
        self.orb["H"] = np.zeros(len(times), dtype="float") + 8
        self.hval = 8

    def test_discovery_metric(self):
        disc_metric = metrics.DiscoveryMetric(
            n_obs_per_night=2,
            t_min=0.0,
            t_max=0.3,
            n_nights_per_window=3,
            t_window=9,
            snr_limit=5,
        )
        metric_value = disc_metric.run(self.sso_obs, self.orb, self.hval)
        child = metrics.DiscoveryNObsMetric(disc_metric)
        n_obs = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        self.assertEqual(n_obs, 8)
        # child = metrics.DiscoveryNObsMetric(disc_metric, i=1)
        # n_obs = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        # self.assertEqual(n_obs, 7)
        child = metrics.DiscoveryNChancesMetric(disc_metric)
        nchances = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        self.assertEqual(nchances, 2)
        child = metrics.DiscoveryTimeMetric(disc_metric)
        time = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        self.assertEqual(time, self.sso_obs["observationStartMJD"][0])
        # child = metrics.DiscoveryTimeMetric(disc_metric, i=1)
        # time = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        # self.assertEqual(time, self.sso_obs["observationStartMJD"][3])
        child = metrics.DiscoveryRadecMetric(disc_metric)
        ra, dec = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        self.assertEqual(ra, 0)
        self.assertEqual(dec, 5)
        child = metrics.DiscoveryEclonlatMetric(disc_metric)
        lon, lat, solar_elong = child.run(self.sso_obs, self.orb, self.hval, metric_value)
        self.assertEqual(lon, 10)
        self.assertEqual(lat, 25)

        disc_metric3 = metrics.MagicDiscoveryMetric(n_obs=5, t_window=2, snr_limit=5)
        magic = disc_metric3.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(magic, 1)
        disc_metric3 = metrics.MagicDiscoveryMetric(n_obs=3, t_window=1, snr_limit=5)
        magic = disc_metric3.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(magic, 2)
        disc_metric3 = metrics.MagicDiscoveryMetric(n_obs=4, t_window=4, snr_limit=5)
        magic = disc_metric3.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(magic, 4)

    def test_high_velocity_metric(self):
        rng = np.random.RandomState(8123)
        vel_metric = metrics.HighVelocityMetric(psf_factor=1.0, snr_limit=5)
        metric_value = vel_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(metric_value, 0)
        self.sso_obs["velocity"][0:2] = 1.5
        metric_value = vel_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(metric_value, 2)
        vel_metric = metrics.HighVelocityMetric(psf_factor=2.0, snr_limit=5)
        metric_value = vel_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(metric_value, 0)
        self.sso_obs["velocity"][0:2] = rng.rand(1)

    def test_high_velocity_nights_metric(self):
        vel_metric = metrics.HighVelocityNightsMetric(psf_factor=1.0, n_obs_per_night=1, snr_limit=5)
        metric_value = vel_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(metric_value, 0)
        self.sso_obs["velocity"][0:2] = 1.5
        metric_value = vel_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(metric_value, self.sso_obs["observationStartMJD"][0])
        self.sso_obs["velocity"][0:2] = np.random.rand(1)


class TestKnownObjectMetrics(unittest.TestCase):
    def setUp(self):
        self.t1 = 53371
        self.t2 = 57023
        self.t3 = 59580
        times = np.arange(self.t1 - 365 * 2, self.t3 + 365 * 3, 1)
        cols = ["MJD(UTC)", "RA", "Dec", "magV", "Elongation", "appMagV"]
        dtype = []
        for c in cols:
            dtype.append((c, "<f8"))
        sso_obs = np.recarray([len(times)], dtype=dtype)

        sso_obs["MJD(UTC)"] = times
        sso_obs["RA"] = np.arange(len(times))
        sso_obs["Dec"] = np.arange(len(times))
        sso_obs["magV"] = np.zeros(len(times), dtype="float") + 20.0
        sso_obs["Elongation"] = np.zeros(len(times), dtype=float) + 180.0
        self.hval = 0.0
        sso_obs["appMagV"] = sso_obs["magV"] + self.hval
        self.orb = None
        self.sso_obs = sso_obs

    def test_known_objects_metric(self):
        known_object_metric = metrics.KnownObjectsMetric(
            t_switch1=self.t1,
            eff1=1.0,
            v_mag_thresh1=20.5,
            t_switch2=self.t2,
            eff2=1.0,
            v_mag_thresh2=20.5,
            t_switch3=self.t3,
            eff3=1.0,
            v_mag_thresh3=20.5,
            eff4=1.0,
            v_mag_thresh4=22,
        )
        m_val = known_object_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(m_val, self.sso_obs["MJD(UTC)"].min())
        known_object_metric = metrics.KnownObjectsMetric(
            t_switch1=self.t1,
            eff1=1.0,
            v_mag_thresh1=15.0,
            t_switch2=self.t2,
            eff2=1.0,
            v_mag_thresh2=20.5,
            t_switch3=self.t3,
            eff3=1.0,
            v_mag_thresh3=20.5,
            eff4=1.0,
            v_mag_thresh4=22,
        )
        m_val = known_object_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(m_val, self.t1)
        known_object_metric = metrics.KnownObjectsMetric(
            t_switch1=self.t1,
            eff1=0.0,
            v_mag_thresh1=20.5,
            t_switch2=self.t2,
            eff2=1.0,
            v_mag_thresh2=20.5,
            t_switch3=self.t3,
            eff3=1.0,
            v_mag_thresh3=20.5,
            eff4=1.0,
            v_mag_thresh4=22,
        )
        m_val = known_object_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(m_val, self.t1)
        known_object_metric = metrics.KnownObjectsMetric(
            t_switch1=self.t1,
            eff1=1.0,
            v_mag_thresh1=10,
            t_switch2=self.t2,
            eff2=1.0,
            v_mag_thresh2=10.5,
            t_switch3=self.t3,
            eff3=1.0,
            v_mag_thresh3=20.5,
            eff4=1.0,
            v_mag_thresh4=22,
        )
        m_val = known_object_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(m_val, self.t2)
        known_object_metric = metrics.KnownObjectsMetric(
            t_switch1=self.t1,
            eff1=1.0,
            v_mag_thresh1=10,
            t_switch2=self.t2,
            eff2=1.0,
            v_mag_thresh2=10.5,
            t_switch3=self.t3,
            eff3=1.0,
            v_mag_thresh3=10.5,
            eff4=1.0,
            v_mag_thresh4=22,
        )
        m_val = known_object_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(m_val, self.t3)
        known_object_metric = metrics.KnownObjectsMetric(
            t_switch1=self.t1,
            eff1=1.0,
            v_mag_thresh1=10,
            t_switch2=self.t2,
            eff2=1.0,
            v_mag_thresh2=10.5,
            t_switch3=self.t3,
            eff3=1.0,
            v_mag_thresh3=10.5,
            eff4=1.0,
            v_mag_thresh4=10.5,
        )
        m_val = known_object_metric.run(self.sso_obs, self.orb, self.hval)
        self.assertEqual(m_val, known_object_metric.badval)


if __name__ == "__main__":
    unittest.main()



"""Tools for use of project-generated opsim simulations and analysis.
"""

__all__ = (
    "get_runs",
    "get_family_runs",
    "download_runs",
    "get_metric_sets",
    "get_metric_summaries",
    "get_family_descriptions",
    "describe_families",
    "create_metric_set_df",
    "write_metric_sets",
)


import copy
import os
import sys
import urllib
import warnings

import numpy as np
import pandas as pd

try:
    import IPython
except ModuleNotFoundError:
    pass

from .summary_plots import plot_run_metric

FAMILY_SOURCE = os.environ.get(
    "RUBIN_SIM_FAMILY_SOURCE",
    "https://raw.githubusercontent.com/lsst-pst/survey_strategy/main/fbs_2.0/runs_v2.2.json",
)

METRIC_SET_SOURCE = os.environ.get(
    "RUBIN_SIM_METRIC_SET_SOURCE",
    "https://raw.githubusercontent.com/lsst-pst/survey_strategy/main/fbs_2.0/metric_sets.json",
)

SUMMARY_SOURCE = os.environ.get(
    "RUBIN_SIM_SUMMARY_SOURCE",
    "https://raw.githubusercontent.com/lsst-pst/survey_strategy/main/fbs_2.0/summary_2023_01_01.csv",
)

if os.uname().nodename.endswith(".datalab.noao.edu"):
    DEFAULT_OPSIM_DB_DIR = "/sims_maf"
else:
    DEFAULT_OPSIM_DB_DIR = os.getcwd()
OPSIM_DB_DIR = os.environ.get("OPSIM_DB_DIR", DEFAULT_OPSIM_DB_DIR)

BY_RUN_COLS = ["run", "brief", "filepath", "url"]


def get_family_runs(run_source=None):
    """Load a data frame that supplies run names for each run family

    Parameters
    ----------
    run_source : `None` or `str`
        File name or URL for the json file from which to load the metadata.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.RUNS_SOURCE` constant.

    Returns
    -------
    families : `pandas.DataFrame`
        ``families``
            The index is the run family. (`str`)
        ``run``
            the project-standard name for the run (`str`)
        ``OpsimGroup``
            The name for the group to which the runs belong (`str`)
        ``OpsimComment``
            Short description of the run (`str`)
        ``OpsimVersion``
            Opsim version name (`str`)
        ``OpsimDate``
            Date for the version of opsim (TODO: ?)
        ``brief``
            A list of descriptions for the run. Runs may have a
            different description for each family it belongs to, so it
            a list of the same length as the families column (`list`
            [`str`])
        ``url``
            The URL from which the opsim output database for this run can be
            downloaded.

    Notes
    -----
    Because runs can be members of multiple families, more than one row may
    provide metadata on the same run.

    The same content (in a different form) can be obtained using
    ``get_runs``. ``get_runs`` is more convenient when indexing by
    run; ``get_family_runs`` when indexing by family.

    """

    run_source = FAMILY_SOURCE if run_source is None else run_source
    if isinstance(run_source, pd.DataFrame):
        runs = run_source
    else:
        families = pd.read_json(run_source, orient="index")
        families.index.name = "family"
        runs = families.explode(BY_RUN_COLS)

    return runs


def get_runs(run_source=None):
    """Load metadata on opsim runs into a `pandas.DataFrame`.

    Parameters
    ----------
    run_source : `None` or `str`
        File name or URL for the json file from which to load the metadata.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.RUNS_SOURCE` constant.

    Returns
    -------
    runs : `pandas.DataFrame`
        ``run``
            The index of the DataFrame is the project-standard name for the run
            (`str`)
        ``family``
            A list of run families to which this run belongs (`list` [`str`])
        ``version``
            The simulation version
        ``brief``
            A list of descriptions for the run. Runs may have a
            different description for each family it belongs to, so it
            a list of the same length as the families column (`list`
            [`str`])
        ``filepath``
            The file path, relative to a base opsim output directory.
        ``url``
            The URL from which the opsim output database for this run can be
            downloaded.

    Notes
    -----
    The same content (in a different form) can be obtained using
    ``get_family_runs``. ``get_runs`` is more convenient when indexing by
    run; ``get_family_runs`` when indexing by family.

    """

    family_runs = get_family_runs(run_source)

    runs = (
        family_runs.reset_index()
        .groupby(BY_RUN_COLS)
        .agg({c: list for c in family_runs.reset_index().columns if c not in BY_RUN_COLS})
        .reset_index()
        .set_index("run")
        .loc[:, ["family", "version", "brief", "filepath", "url"]]
    )

    return runs


def download_runs(runs, dest_dir=None, runs_source=None, clobber=False):
    """Download opsim visit databases for specified runs to a local directory.

    Parameters
    ----------
    runs : `pandas.DataFrame` or iterable [`str`]
        If a `pandas.DataFrame` is provided, the `OpsimRun` column will be used
        to get run names, and data will be read from the url specified in the
        `url` column.
        If a collection of `str` is provided, these will be interpreted as
        run names supplied by data originating in the run metadata provided
        by the ``runs_source`` parameter.
    dest_dir : `str`
        The local directory into which to write downloaded visit databases.
    runs_source : `str`
        File name or URL for the json file from which to load the metadata.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.RUNS_SOURCE` constant. This parameter is ignored
        if the ``runs`` parameter is set to a `pandas.DataFrame`.
    clobber : `bool`
        If ``False``, runs that would clobber an existing file will be skipped.
        If ``True``, existing files will be overwritten.


    Returns
    -------
    runs : `pandas.DataFrame`
        Metadata on runs downloaded (in the same structure as the return of
        ``archive.get_runs``).
    """

    if isinstance(runs, str):
        runs = [runs]

    if not isinstance(runs, pd.DataFrame):
        all_runs = get_runs(runs_source)
        runs = all_runs.loc[runs, :]

    if dest_dir is None:
        dest_dir = OPSIM_DB_DIR

    if not os.path.exists(dest_dir):
        raise FileNotFoundError(dest_dir)

    dest_fnames = pd.Series(name="fname", index=pd.Index([], name="OpsimRun"), dtype=object)

    for run_name, run in runs.iterrows():
        dest_fnames[run_name] = os.path.join(dest_dir, run.filepath)

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(dest_fnames[run_name]), exist_ok=True)

        if clobber or not os.path.exists(dest_fnames[run_name]):
            urllib.request.urlretrieve(run.url, dest_fnames[run_name])
        else:
            warnings.warn(f"{dest_fnames[run_name]} already exists; not downloading")

    return dest_fnames


def get_metric_sets(metric_set_source=METRIC_SET_SOURCE):
    """Get metadata on named sets of related metrics.

    Parameters
    ----------
    metric_set_source : `str`
        File name or URL for the json file from which to load the data.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.METRIC_SET_SOURCE` constant.

    Returns
    -------
    metric_sets : `pandas.DataFrame`
        ``metric_set``
            The 1st level of the index is the name of a set of metrics (`str`).
        ``metric``
            The 2nd level of the index is the full name of the metric (`str`).
        ``metric``
            The full name of the metric (`str`).
        ``short_name``
            An abbreviated name for the metric (`str`)..
        ``style``
            The ``matplotlib`` linestyle suggested for plots of the
            metric (`str`).
        ``invert``
            When normalizing, invert the metric value first? (`bool`)
        ``mag``
            Is the value an (astronomical) magnitude? (`bool`)
    """
    metric_set_source = METRIC_SET_SOURCE if metric_set_source is None else metric_set_source
    if isinstance(metric_set_source, pd.DataFrame):
        metric_sets = metric_set_source
    else:
        metric_sets = (
            pd.read_json(metric_set_source)
            .set_index("metric set")
            .set_index("metric", append=True, drop=False)
        )
    return metric_sets


def write_metric_sets(metric_set_file, metric_sets):
    """Write an updated metric_set dataframe to disk

    Parameters
    ----------
    metric_set_file : `str`
        Output file name.
    metric_sets : `pandas.DataFrame`
        Metric_set dataframe, as defined in get_metric_sets
    """
    tmp = metric_sets.reset_index("metric set")
    tmp.to_json(metric_set_file, orient="records", indent=2)


def get_metric_summaries(
    run_families=tuple(),
    metric_sets=tuple(),
    runs=tuple(),
    metrics=tuple(),
    summary_source=None,
    runs_source=None,
    metric_set_source=None,
    run_order="family",
    metric_order="summary",
):
    """Get summary metric values for a set of runs and metrics.

    Parameters
    ----------
    run_families : iterable [`str`]
        Families of runs to include in the summary.
    metric_sets : iterable [`str`]
        Sets of metrics to include in the summary.
    runs : iterable [`str`]
        Runs to include in the summary (in addition to any that are part
        of families included in ``run_families``).
    metrics : iterable [`str`]
        Metrics to include in the summary (in addition to any that are
        part of sets included in ``metric_sets``).
    summary_source : `str` or `pandas.DataFrame`
        File name or URL for the file from which to load the data.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.METRIC_SET_SOURCE` constant.
        If the supplied value is a `pandas.DataFrame`, it the table
        returned will be a subset of this supplied table.
    run_source : `pandas.DataFrame` or `str`
        Either a `pandas.DataFrame` of runs metadata (as returned by
        `archive.get_runs`), or a file name or URL for the json file
        from which to load the run metadata.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.RUNS_SOURCE` constant.
    metric_set_source : `pandas.DataFrame` or `str`
        Either a `pandas.DataFrame` of metric set specifications
        (as returned by `archive.get_metric_sets`) or a
        file name or URL for the json file from which to load the data.
        If it is set to `None`, the data is loaded from the URL specified
        by the `archive.SUMMARY_SOURCE` constant.
    run_order : `str`
        Sort runs according to family definition ("family") or summary file
        ("summary") order.
    metric_order : `str`
        Sort metrics according to set definition ("set") or summary file
        ("summary") order.

    Returns
    -------
    summaries : `pandas.DataFrame`
        Metric summary values are returned in a `pandas.DataFrame`, with
        each column providing the metrics for one run, and each row the
        values for one metric. The metric names constitute the index, and
        the column names are the canonical run names.

    Note
    ----
    The entire summary statistic values for all of the runs and metrics can
    be downloaded from the default sources first, by simply calling
    .. code-block:: python
        summary = get_metric_summaries()

    Then, you can use `get_metric_summaries` to get a subset without
    redownloading the whole set by passing `summary_source=summary`. If you are
    selecting multiple subsets of the summary, this avoids needing
    to download the summary data multiple times.
    """
    summary_source = SUMMARY_SOURCE if summary_source is None else summary_source

    runs = list(runs)
    metrics = list(metrics)

    if isinstance(run_families, str):
        run_families = [run_families]

    if isinstance(metric_sets, str):
        metric_sets = [metric_sets]

    if isinstance(summary_source, pd.DataFrame):
        all_summaries = summary_source
    else:
        try:
            all_summaries = pd.read_csv(summary_source, index_col=0, low_memory=False)
        except UnicodeDecodeError:
            # then this was probably the h5 file instead
            all_summaries = pd.read_hdf(summary_source)
        all_summaries.index.name = "OpsimRun"

    if len(run_families) > 0:
        families = get_family_runs(runs_source)
        for run_family in run_families:
            runs.extend(pd.Series(families.loc[run_family, "run"]).tolist())

    if len(metric_sets) > 0:
        metric_set_df = get_metric_sets(metric_set_source)
        for metric_set in metric_sets:
            metrics.extend(list(metric_set_df.loc[metric_set, "metric"]))

    if len(runs) == 0:
        runs = slice(None)
    else:
        if run_order == "summary":
            runs = [r for r in all_summaries.index if r in runs]

    if len(metrics) == 0:
        metrics = slice(None)
    else:
        requested_metrics = copy.copy(metrics)
        for metric in requested_metrics:
            if metric not in all_summaries.columns:
                warnings.warn(f'Metric "{metric}" not in summary, skipping')
                metrics.remove(metric)

        if metric_order == "summary":
            metrics = [m for m in all_summaries.columns if m in metrics]

    summaries = all_summaries.loc[runs, metrics]
    summaries.columns.name = "metric"
    summaries.index.name = "run"
    return summaries


def get_family_descriptions(family_source=None):
    """Get description of families or funs.

    Parameters
    ----------
    family_source : `str
        File name or URL for the json file from which to load the
        family descriptinos.  If it is set to `None`, the data is
        loaded from the URL specified by the
        `archive.FAMILY_SOURCE` constant.


    Returns
    -------
    families : `pandas.DataFrame`
        Family descriptions, with comments.

    """
    family_source = FAMILY_SOURCE if family_source is None else family_source
    if isinstance(family_source, pd.DataFrame):
        families = family_source
    else:
        families = pd.read_json(family_source, orient="index")
        families.index.name = "family"
        by_family_cols = [c for c in families.columns if c not in BY_RUN_COLS]
        families = families.loc[:, by_family_cols + BY_RUN_COLS]
    return families


def describe_families(
    families,
    summary=None,
    table_metric_set=None,
    plot_metric_set=None,
    baseline_run=None,
    round_table=2,
):
    """Display (in a jupyter on IPython notebook) family descirptions

    Parameters
    ----------
    families : `pandas.DataFrame`
        Data family descriptions as returned by get_family_descriptions.
    summary : `pandas.DataFrame`
        Summary metrics for each run, as returned by get_metric_summaries.
    table_metric_set : `pandas.DataFrame`
        Metadata on metrics to be included in the table, with columns and
        index as returned by get_metric_sets. None if no metrics should be
        included in the table.
    plot_metric_set : `pandas.DataFrame`
        Metadata on metrics to be included in the plot, with columns and
        index as returned by get_metric_sets. None if no plot should be
        made.
    baseline_run : `str`
        The name of the run to use to normalize metrics in the plot.
        None if normalization should be skipped.
    round_table : `int`, opt
        Decimal places to which to round the table_metrics. Default 2.

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        The plot figure.
    ax : `matplotilb.axes.Axes`
        The plot axes.
    """

    family_runs = families.explode(["run", "brief", "filepath"]).loc[:, ["run", "brief", "filepath"]]

    # If there is just one run in the family, we might
    # get a pd.Series back rather than a pd.DataFrame.
    # Make sure we have a DataFrame
    if isinstance(family_runs, pd.Series):
        family_runs = pd.DataFrame([family_runs])

    for family_name, family in families.iterrows():
        # Use awkward appending of each line to string rather
        # than a tripple quote to keep flake8 from complaining
        # about blanks at the end of lines, which are
        # meaningful in markdown (and desired here).
        description = "---\n"
        description += f"{family.description}  \n"
        description += f"**version**: {family.version}  \n"
        description += "**runs**:  \n"

        these_runs = family_runs.loc[[family_name], :]
        if summary is not None:
            if table_metric_set is not None:
                table_metric_summary = summary.loc[these_runs["run"], table_metric_set["metric"]]
                table_metric_summary.rename(table_metric_set["short_name"], axis=1, inplace=True)
                if round_table is not None:
                    table_metric_summary = table_metric_summary.round(round_table)
            else:
                table_metric_summary = summary.loc[these_runs["run"]]

            these_runs = these_runs.join(table_metric_summary, on="run", how="left")

            num_columns = len(these_runs.columns)
            if num_columns > 5 and "filepath" in these_runs.columns:
                these_runs = these_runs.drop(columns=["filepath"])

        with pd.option_context("display.max_colwidth", 0):
            if "IPython" in sys.modules:
                IPython.display.display_markdown(description, raw=True)
                IPython.display.display(
                    IPython.display.HTML(these_runs.set_index("run").to_html().replace("\\n", "<br>"))
                )
            else:
                print(description)
                print(these_runs.set_index("run"))

    if plot_metric_set is not None:
        these_runs = family_runs["run"].values
        if baseline_run is not None and baseline_run not in these_runs:
            these_runs = np.concatenate([[baseline_run], these_runs])
        these_metrics = [m for m in plot_metric_set["metric"] if m in summary.columns]
        fig, ax = plot_run_metric(  # pylint: disable=invalid-name
            summary.loc[these_runs, these_metrics],
            metric_set=plot_metric_set,
            metric_label_map=plot_metric_set["short_name"],
            baseline_run=baseline_run,
            vertical_quantity="value",
            horizontal_quantity="run",
        )
    else:
        fig, ax = None, None  # pylint: disable=invalid-name

    return fig, ax


def create_metric_set_df(
    metric_set,
    metrics,
    short_name=None,
    style="-",
    invert=False,
    mag=False,
):
    """Create a DataFrame that defines a metric set.

    Parameters
    ----------
    metric_set : `str`
        The name of a metric set.
    metrics : `list` [`str`]
        A list of metric names in the set.
    short_name : `list` [`str`], optional
        A list of shorter metric names, by default None
    style : `list` [`str`], optional
        The matplotlib line style symbol for lines representing the metric,
        by default "-"
    invert : `list` [`bool`], optional
        Are smaller values of the metric better, such as for errors?, by default False
    mag : `list` [`bool`], optional
        Is the metric an astronomical magnitude?, by default False

    Returns
    -------
    metric_set : `pandas.DataFrame`
        A table of metrics and normalization and plotting flags defining the
        content of a metric set.
    """
    if short_name is None:
        short_name = metrics

    metric_set = (
        pd.DataFrame(
            {
                "metric set": metric_set,
                "metric": metrics,
                "short_name": short_name,
                "style": style,
                "invert": invert,
                "mag": mag,
            }
        )
        .set_index("metric set")
        .set_index("metric", append=True, drop=False)
    )

    return metric_set



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestTechnicalMetrics(unittest.TestCase):
    def test_n_changes_metric(self):
        """
        Test the NChanges metric.
        """
        filters = np.array(["u", "u", "g", "g", "r"])
        visit_times = np.arange(0, filters.size, 1)
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.NChangesMetric()
        result = metric.run(data)
        self.assertEqual(result, 2)
        filters = np.array(["u", "g", "u", "g", "r"])
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.NChangesMetric()
        result = metric.run(data)
        self.assertEqual(result, 4)

    def test_min_time_between_states_metric(self):
        """
        Test the minTimeBetweenStates metric.
        """
        filters = np.array(["u", "g", "g", "r"])
        visit_times = np.array([0, 5, 6, 7])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MinTimeBetweenStatesMetric()
        result = metric.run(data)  # minutes
        self.assertEqual(result, 2 * 24.0 * 60.0)
        data["filter"] = np.array(["u", "u", "u", "u"])
        result = metric.run(data)
        self.assertEqual(result, metric.badval)

    def test_n_state_changes_faster_than_metric(self):
        """
        Test the NStateChangesFasterThan metric.
        """
        filters = np.array(["u", "g", "g", "r"])
        visit_times = np.array([0, 5, 6, 7])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.NStateChangesFasterThanMetric(cutoff=3 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 1)

    def test_max_state_changes_within_metric(self):
        """
        Test the MaxStateChangesWithin metric.
        """
        filters = np.array(["u", "g", "r", "u", "g", "r"])
        visit_times = np.array([0, 1, 1, 4, 6, 7])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 2)
        filters = np.array(["u", "g", "g", "u", "g", "r", "g", "r"])
        visit_times = np.array([0, 1, 1, 4, 4, 7, 8, 8])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 3)

        filters = np.array(["u", "g"])
        visit_times = np.array([0, 1])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 1)

        filters = np.array(["u", "u"])
        visit_times = np.array([0, 1])  # days
        data = np.core.records.fromarrays([visit_times, filters], names=["observationStartMJD", "filter"])
        metric = metrics.MaxStateChangesWithinMetric(timespan=1 * 24 * 60)
        result = metric.run(data)  # minutes
        self.assertEqual(result, 0)

    def test_teff_metric(self):
        """
        Test the Teff (time_effective) metric.
        """
        filters = np.array(["g", "g", "g", "g", "g"])
        m5 = np.zeros(len(filters), float) + 25.0
        data = np.core.records.fromarrays([m5, filters], names=["fiveSigmaDepth", "filter"])
        metric = metrics.TeffMetric(fiducial_depth={"g": 25}, teff_base=30.0)
        result = metric.run(data)
        self.assertEqual(result, 30.0 * m5.size)
        filters = np.array(["g", "g", "g", "u", "u"])
        m5 = np.zeros(len(filters), float) + 25.0
        m5[3:5] = 20.0
        data = np.core.records.fromarrays([m5, filters], names=["fiveSigmaDepth", "filter"])
        metric = metrics.TeffMetric(fiducial_depth={"u": 20, "g": 25}, teff_base=30.0)
        result = metric.run(data)
        self.assertEqual(result, 30.0 * m5.size)

    def test_open_shutter_fraction_metric(self):
        """
        Test the open shutter fraction metric.
        """
        nvisit = 10
        exptime = 30.0
        slewtime = 30.0
        visit_exp_time = np.ones(nvisit, dtype="float") * exptime
        visit_time = np.ones(nvisit, dtype="float") * (exptime + 0.0)
        slew_time = np.ones(nvisit, dtype="float") * slewtime
        data = np.core.records.fromarrays(
            [visit_exp_time, visit_time, slew_time],
            names=["visitExposureTime", "visitTime", "slewTime"],
        )
        metric = metrics.OpenShutterFractionMetric()
        result = metric.run(data)
        self.assertEqual(result, 0.5)

    def test_brute_osf_metric(self):
        """
        Test the open shutter fraction metric.
        """
        nvisit = 10
        exptime = 30.0
        slewtime = 30.0
        visit_exp_time = np.ones(nvisit, dtype="float") * exptime
        visit_time = np.ones(nvisit, dtype="float") * (exptime + 0.0)
        slew_time = np.ones(nvisit, dtype="float") * slewtime
        mjd = np.zeros(nvisit) + np.add.accumulate(visit_exp_time) + np.add.accumulate(slew_time)
        mjd = mjd / 60.0 / 60.0 / 24.0
        data = np.core.records.fromarrays(
            [visit_exp_time, visit_time, slew_time, mjd],
            names=[
                "visitExposureTime",
                "visit_time",
                "slew_time",
                "observationStartMJD",
            ],
        )
        metric = metrics.BruteOSFMetric()
        result = metric.run(data)
        self.assertGreater(result, 0.5)
        self.assertLess(result, 0.6)


if __name__ == "__main__":
    unittest.main()



__all__ = ("ddfBatch",)

import healpy as hp
import numpy as np
from rubin_scheduler.utils import angular_separation, ddf_locations, hpid2_ra_dec, sample_patch_on_sphere

import rubin_sim.maf as maf

from .common import lightcurve_summary


def ddfBatch(
    run_name="run_name",
    nside=512,
    radius=2.5,
    nside_sne=128,
    extra_sql=None,
    extra_info_label=None,
):
    """
    A set of metrics to evaluate DDF fields.

    Parameters
    ----------
    run_name : `str`, optional
        The name of the simulation (for plot titles and file outputs).
    nside : `int`, optional
        The HEALpix nside to run most of the metrics on.
    radius : `float`
        The radius to select around each ddf (degrees).
        The default value of 2.5 degrees has been chosen to balance
        selecting a large enough area to ensure gathering all of the double
        Euclid field or a run with large dithers,
        while not including too much background area (which can
        skew metrics of the median number of visits, etc.).
    nside_sne : `int`, optional
        The HEALpix nside to use with the SNe metric. The default is lower
        than the default nside for other metrics, as the SNe metric is
        more computationally expensive.
    extra_sql : `str`, optional
        Additional sql constraint (such as night<=365) to add to the
        necessary sql constraints for each metric.
    extra_info_label : `str`, optional
        Additional description information to add (alongside the extra_sql)

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    bundle_list = []

    # Define the slicer to use for each DDF
    # Get standard DDF locations and reformat information as a dictionary
    ddfs = {}
    ddfs_rough = ddf_locations()
    for ddf in ddfs_rough:
        ddfs[ddf] = {"ra": ddfs_rough[ddf][0], "dec": ddfs_rough[ddf][1]}
    # Combine the Euclid double-field into one - but with two ra/dec values
    ddfs["EDFS"] = {
        "ra": [ddfs["EDFS_a"]["ra"], ddfs["EDFS_b"]["ra"]],
        "dec": [ddfs["EDFS_a"]["dec"], ddfs["EDFS_b"]["dec"]],
    }
    del ddfs["EDFS_a"]
    del ddfs["EDFS_b"]
    # Let's include an arbitrary point that should be in the WFD for comparison
    ddfs["WFD"] = {"ra": 0, "dec": -20.0}

    ra, dec = hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
    ra_sne, dec_sne = hpid2_ra_dec(nside_sne, np.arange(hp.nside2npix(nside_sne)))

    ddf_slicers = {}
    ddf_slicers_sne = {}
    for ddf in ddfs:
        # Define the healpixels to use for this DDF
        if np.size(ddfs[ddf]["ra"]) > 1:
            goods = []
            goods_sne = []
            for ddf_ra, ddf_dec in zip(ddfs[ddf]["ra"], ddfs[ddf]["dec"]):
                dist = angular_separation(ra, dec, ddf_ra, ddf_dec)
                goods.append(np.where(dist <= radius)[0])
                dist = angular_separation(ra_sne, dec_sne, ddf_ra, ddf_dec)
                goods_sne.append(np.where(dist <= radius)[0])
            good = np.unique(np.concatenate(goods))
            good_sne = np.unique(np.concatenate(goods_sne))
        else:
            dist = angular_separation(ra, dec, np.mean(ddfs[ddf]["ra"]), np.mean(ddfs[ddf]["dec"]))
            good = np.where(dist <= radius)[0]
            dist = angular_separation(ra_sne, dec_sne, np.mean(ddfs[ddf]["ra"]), np.mean(ddfs[ddf]["dec"]))
            good_sne = np.where(dist <= radius)[0]
        ddf_slicers_sne[ddf] = maf.HealpixSubsetSlicer(nside_sne, good_sne, use_cache=False)
        ddf_slicers[ddf] = maf.HealpixSubsetSlicer(nside, good, use_cache=False)

    # Now define metrics

    # Set up basic all and per filter sql constraints.
    filterlist, colors, orders, sqls, info_labels = maf.filter_list(
        all=True, extra_sql=extra_sql, extra_info_label=extra_info_label
    )

    summary_stats = [maf.MeanMetric(), maf.MedianMetric(), maf.SumMetric()]
    depth_stats = [maf.MedianMetric()]

    plotFuncs = [maf.HealpixSkyMap()]

    displayDict = {"group": "", "subgroup": "", "order": 0}

    order = 0
    for ddf in ddfs:
        fieldname = ddf
        if not (fieldname.startswith("DD")):
            fieldname = f"DD:{fieldname}"

        plotDict = {
            "visufunc": hp.gnomview,
            "rot": (np.mean(ddfs[ddf]["ra"]), np.mean(ddfs[ddf]["dec"]), 0),
            "xsize": 500,
        }
        order += 1

        # Number of SNe
        displayDict["group"] = "SNe"
        displayDict["subgroup"] = "N SNe"
        displayDict["caption"] = f"SNIa in the {fieldname} DDF."
        displayDict["order"] = order

        metric = maf.metrics.SNNSNMetric(
            verbose=False,
            n_bef=4,
            n_aft=10,
            zmin=0.1,
            zmax=1.1,
            z_step=0.03,
            daymax_step=3,
            coadd_night=True,
            gamma_name="gamma_DDF.hdf5",
            # have to add field name here, to avoid reduceDict key collissions
            metric_name=f"SNNSNMetric {fieldname}",
        )
        bundle_list.append(
            maf.MetricBundle(
                metric,
                ddf_slicers_sne[ddf],
                constraint="",
                info_label=" ".join([fieldname, "all bands, only DDF observations"]),
                plot_dict=plotDict,
                plot_funcs=plotFuncs,
                summary_metrics=summary_stats,
                display_dict=displayDict,
            )
        )
        # Strong lensed SNe
        displayDict["group"] = "SNe"
        displayDict["subgroup"] = "SL SNe"
        displayDict["caption"] = f"Strongly Lensed SN metric in the {fieldname} DDF."
        displayDict["order"] = order
        metric = maf.SNSLMetric()
        bundle_list.append(
            maf.MetricBundle(
                metric,
                ddf_slicers[ddf],
                constraint=sqls["all"],
                info_label=" ".join([fieldname, info_labels["all"]]),
                plot_dict=plotDict,
                plot_funcs=plotFuncs,
                summary_metrics=summary_stats,
                display_dict=displayDict,
            )
        )

        # KNe

        delta = 5.0  # degrees
        n_kne = 5000
        displayDict["group"] = "KNe"
        displayDict["subgroup"] = ""
        displayDict["caption"] = f"Number of KNe in the {fieldname} DDF from %i injected." % n_kne

        ra, dec = sample_patch_on_sphere(
            np.mean(ddfs[ddf]["ra"]), np.mean(ddfs[ddf]["dec"]), delta, n_kne, seed=1
        )

        metric = maf.KNePopMetric(metric_name="KNePopMetric_%s" % fieldname)
        slicer = maf.generate_kn_pop_slicer(n_events=n_kne, ra=ra, dec=dec)

        bundle_list.append(
            maf.MetricBundle(
                metric,
                slicer,
                "",
                info_label=" ".join([fieldname]),
                summary_metrics=lightcurve_summary(),
                display_dict=displayDict,
            )
        )

        # kuiper metric
        displayDict["group"] = "Kuiper"
        displayDict["subgroup"] = ""
        displayDict["caption"] = f"Kuiper metric in the {fieldname} DDF."

        sqls_gri = {
            "gri": "filter='g' or filter='r' or filter='i'",
            "riz": "filter='r' or filter='i' or filter='z'",
        }

        for sql in sqls_gri:
            metrics = [
                maf.KuiperMetric(
                    "rotSkyPos",
                    metric_name=f"Kuiper statistic (0 is uniform, 1 is delta function),rotSkyPos,{fieldname},"
                    + sql,
                ),
                maf.KuiperMetric(
                    "rotTelPos",
                    metric_name=f"Kuiper statistic (0 is uniform, 1 is delta function),rotTelPos,{fieldname},"
                    + sql,
                ),
            ]
            for metric in metrics:
                bundle_list.append(
                    maf.MetricBundle(
                        metric,
                        ddf_slicers[ddf],
                        sqls_gri[sql],
                        info_label=" ".join([fieldname]),
                        plot_dict=plotDict,
                        plot_funcs=plotFuncs,
                        summary_metrics=summary_stats,
                        display_dict=displayDict,
                    )
                )

        # Weak lensing visits
        # The "magic numbers" here scale the final depth into
        # approximately consistent visits per year - final depth is
        # determined by arbitrary definition of 'good sample'
        lim_ebv = 0.2
        offset = 0.1
        mag_cuts = {
            1: 24.75 - offset,
            2: 25.12 - offset,
            3: 25.35 - offset,
            4: 25.5 - offset,
            5: 25.62 - offset,
            6: 25.72 - offset,
            7: 25.8 - offset,
            8: 25.87 - offset,
            9: 25.94 - offset,
            10: 26.0 - offset,
        }
        displayDict["group"] = "Weak Lensing"
        displayDict["subgroup"] = ""
        displayDict["caption"] = f"Weak lensing metric in the {fieldname} DDF."

        sqls_gri = {
            "gri": "filter='g' or filter='r' or filter='i'",
            "riz": "filter='r' or filter='i' or filter='z'",
        }

        for sql in sqls_gri:
            metric = maf.WeakLensingNvisits(
                lsst_filter="i",
                depth_cut=mag_cuts[10],
                ebvlim=lim_ebv,
                min_exp_time=20.0,
                metric_name="WeakLensingNvisits_" + sql,
            )
            bundle_list.append(
                maf.MetricBundle(
                    metric,
                    ddf_slicers[ddf],
                    sqls_gri[sql],
                    info_label=" ".join([fieldname, sql]),
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    summary_metrics=summary_stats,
                    display_dict=displayDict,
                )
            )

        # Number of QSOs in each band
        displayDict["group"] = "QSO"
        displayDict["subgroup"] = "Number"
        displayDict["caption"] = f"Number of QSO in the {fieldname} DDF."
        zmin = 0.3
        extinction_cut = 1.0
        for f in "ugrizy":
            displayDict["order"] = orders[f]
            summaryMetrics = [maf.SumMetric(metric_name="Total QSO")]
            metric = maf.QSONumberCountsMetric(
                f,
                units="mag",
                extinction_cut=extinction_cut,
                qlf_module="Shen20",
                qlf_model="A",
                sed_model="Richards06",
                zmin=zmin,
                zmax=None,
            )
            bundle_list.append(
                maf.MetricBundle(
                    metric,
                    ddf_slicers[ddf],
                    sqls[f],
                    info_label=" ".join([fieldname, info_labels[f]]),
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    summary_metrics=summaryMetrics,
                    display_dict=displayDict,
                )
            )

        # Run the TimeLag for each filter *and* all filters
        displayDict["group"] = "QSO"
        displayDict["subgroup"] = "TimeLags"
        nquist_threshold = 2.2
        lag = 100
        summaryMetrics = [maf.MeanMetric(), maf.MedianMetric(), maf.RmsMetric()]
        m = maf.AgnTimeLagMetric(threshold=nquist_threshold, lag=lag)
        for f in filterlist:
            displayDict["order"] = orders[f]
            displayDict["caption"] = (
                f"Comparion of the time between visits compared to a defined sampling gap ({lag} days) in "
                f"{f} band."
            )
            bundle_list.append(
                maf.MetricBundle(
                    m,
                    ddf_slicers[ddf],
                    constraint=sqls[f],
                    info_label=" ".join([fieldname, info_labels[f]]),
                    run_name=run_name,
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    summary_metrics=summaryMetrics,
                    display_dict=displayDict,
                )
            )

        # Run the TimeLag for each filter *and* all filters
        displayDict["group"] = "QSO"
        displayDict["subgroup"] = "TimeLags"
        nquist_threshold = 2.2
        lag = 5
        summaryMetrics = [maf.MeanMetric(), maf.MedianMetric(), maf.RmsMetric()]
        m = maf.AgnTimeLagMetric(threshold=nquist_threshold, lag=lag)
        for f in filterlist:
            displayDict["order"] = orders[f]
            displayDict["caption"] = (
                f"Comparion of the time between visits compared to a defined sampling gap ({lag} days) in "
                f"{f} band."
            )
            bundle_list.append(
                maf.MetricBundle(
                    m,
                    ddf_slicers[ddf],
                    constraint=sqls[f],
                    info_label=" ".join([fieldname, info_labels[f]]),
                    run_name=run_name,
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    summary_metrics=summaryMetrics,
                    display_dict=displayDict,
                )
            )

        # AGN structure function
        displayDict["group"] = "QSO"
        displayDict["subgroup"] = "Structure Function"
        agn_mags = {"u": 22.0, "g": 24, "r": 24, "i": 24, "z": 22, "y": 22}
        for f in "ugrizy":
            displayDict["order"] = orders[f]
            displayDict["caption"] = f"AGN Structure Function Error in {f} band in the {fieldname} DDF."
            summaryMetrics = [maf.MedianMetric(), maf.RmsMetric()]
            metric = maf.SFUncertMetric(
                mag=agn_mags[f],
                bins=np.logspace(0, np.log10(3650), 21),
            )
            bundle_list.append(
                maf.MetricBundle(
                    metric,
                    ddf_slicers[ddf],
                    sqls[f],
                    info_label=" ".join([fieldname, info_labels[f]]),
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    summary_metrics=summaryMetrics,
                    display_dict=displayDict,
                )
            )

        # Coadded depth per filter, and count per filter
        displayDict["group"] = "Basics"
        for f in "ugrizy":
            displayDict["subgroup"] = "Coadd M5"
            displayDict["order"] = orders[f]
            displayDict["caption"] = f"Coadded m5 in {f} band in the {fieldname} DDF."
            metric = maf.Coaddm5Metric(metric_name=f"{fieldname} CoaddM5")
            bundle_list.append(
                maf.MetricBundle(
                    metric,
                    ddf_slicers[ddf],
                    sqls[f],
                    info_label=info_labels[f],
                    plot_dict=plotDict,
                    run_name=run_name,
                    plot_funcs=plotFuncs,
                    summary_metrics=depth_stats,
                    display_dict=displayDict,
                )
            )
            displayDict["subgroup"] = "N Visits"
            displayDict["caption"] = f"Number of visits in the {f} band in the {fieldname} DDF."
            metric = maf.CountMetric(col="observationStartMJD", units="#", metric_name=f"{fieldname} NVisits")
            bundle_list.append(
                maf.MetricBundle(
                    metric,
                    ddf_slicers[ddf],
                    sqls[f],
                    info_label=info_labels[f],
                    plot_dict=plotDict,
                    run_name=run_name,
                    plot_funcs=plotFuncs,
                    summary_metrics=depth_stats,
                    display_dict=displayDict,
                )
            )
        # Count over all filter
        displayDict["subgroup"] = "N Visits"
        displayDict["order"] = orders["all"]
        displayDict["caption"] = f"Number of visits in all bands in the {fieldname} DDF."
        metric = maf.CountMetric(col="observationStartMJD", units="#", metric_name=f"{fieldname} NVisits")
        bundle_list.append(
            maf.MetricBundle(
                metric,
                ddf_slicers[ddf],
                constraint=sqls["all"],
                info_label=info_labels["all"],
                plot_dict=plotDict,
                run_name=run_name,
                plot_funcs=plotFuncs,
                summary_metrics=depth_stats,
                display_dict=displayDict,
            )
        )
        # Count number of unique nights with visits
        displayDict["group"] = "Cadence"
        displayDict["subgroup"] = "N Nights"
        displayDict["order"] = orders["all"]
        displayDict["caption"] = f"Number of nights with visits in the {fieldname} DDF."
        metric = maf.CountUniqueMetric(col="night", units="#", metric_name=f"{fieldname} N Unique Nights")
        bundle_list.append(
            maf.MetricBundle(
                metric,
                ddf_slicers[ddf],
                constraint=sqls["all"],
                info_label=info_labels["all"],
                plot_dict=plotDict,
                run_name=run_name,
                plot_funcs=plotFuncs,
                summary_metrics=depth_stats,
                display_dict=displayDict,
            )
        )

        # Now to compute some things at just the center of the DDF
        # For these metrics, add a requirement that the 'note' label
        # match the DDF, to avoid WFD visits skewing the results
        # (we want to exclude non-DD visits),
        ptslicer = maf.UserPointsSlicer(np.mean(ddfs[ddf]["ra"]), np.mean(ddfs[ddf]["dec"]))

        displayDict["group"] = "Cadence"
        displayDict["order"] = order

        fieldsqls = {}
        if ddf == "WFD":
            for f in filterlist:
                fieldsqls[f] = sqls[f]
        else:
            fieldsql = f"note like '%{fieldname}%'"
            for f in filterlist:
                if len(sqls[f]) > 0:
                    fieldsqls[f] = fieldsql + " and " + sqls[f]
                else:
                    fieldsqls[f] = fieldsql

        displayDict["subgroup"] = "Sequence length"

        # Number of observations per night, any filter (sequence length)
        # Histogram the number of visits per night at the center of the DDF
        countbins = np.arange(0, 200, 5)
        metric = maf.NVisitsPerNightMetric(
            night_col="night",
            bins=countbins,
            metric_name=f"{fieldname} NVisitsPerNight",
        )
        plotDict = {"bins": countbins, "xlabel": "Number of visits per night"}
        displayDict[
            "caption"
        ] = f"Histogram of the number of visits in each night, at the center of {fieldname}."
        plotFunc = maf.SummaryHistogram()
        bundle = maf.MetricBundle(
            metric,
            ptslicer,
            fieldsqls["all"],
            info_label=info_labels["all"],
            plot_dict=plotDict,
            display_dict=displayDict,
            plot_funcs=[plotFunc],
        )
        bundle_list.append(bundle)

        if fieldname.endswith("WFD"):
            pass
        else:
            displayDict["caption"] = f"Number of visits per night for {fieldname}."
            metric = maf.CountMetric("observationStartMJD", metric_name=f"{fieldname} Nvisits Per Night")
            slicer = maf.OneDSlicer(slice_col_name="night", bin_size=1)
            bundle = maf.MetricBundle(
                metric,
                slicer,
                fieldsqls["all"],
                info_label=info_labels["all"],
                display_dict=displayDict,
                summary_metrics=[
                    maf.MedianMetric(),
                    maf.PercentileMetric(percentile=80, metric_name="80thPercentile"),
                    maf.MinMetric(),
                    maf.MaxMetric(),
                ],
            )
            bundle_list.append(bundle)

        displayDict["subgroup"] = "Sequence gaps"

        # Histogram of the number of nights between visits, all filters
        bins = np.arange(1, 40, 1)
        metric = maf.NightgapsMetric(
            bins=bins,
            night_col="night",
            metric_name=f"{fieldname} Delta Nights Histogram",
        )
        displayDict["caption"] = f"Histogram of intervals between nights with visits, in the {fieldname} DDF."
        plotDict = {"bins": bins, "xlabel": "dT (nights)"}
        plotFunc = maf.SummaryHistogram()
        bundle = maf.MetricBundle(
            metric,
            ptslicer,
            constraint=fieldsqls["all"],
            info_label=info_labels["all"],
            plot_dict=plotDict,
            display_dict=displayDict,
            plot_funcs=[plotFunc],
        )
        bundle_list.append(bundle)

        # Median inter-night gap in each and all filters
        for f in filterlist:
            metric = maf.InterNightGapsMetric(
                metric_name=f"{fieldname} Median Inter-Night Gap", reduce_func=np.median
            )
            displayDict["order"] = orders[f]
            displayDict["caption"] = f"Median internight gap in {f} band in the {fieldname} DDF."
            bundle_list.append(
                maf.MetricBundle(
                    metric,
                    ptslicer,
                    fieldsqls[f],
                    info_label=info_labels[f],
                    run_name=run_name,
                    summary_metrics=[maf.MeanMetric()],
                    plot_funcs=[],
                    display_dict=displayDict,
                )
            )

        displayDict["subgroup"] = "Season length"

        # Histogram of the season lengths, all filters
        def rfunc(simdata):
            # Sometimes number of seasons is 10, sometimes 11
            # (depending on where survey starts/end)
            # so normalize it so there's always 11 values
            if len(simdata) < 11:
                simdata = np.concatenate([simdata, np.array([0], float)])
            return simdata

        metric = maf.SeasonLengthMetric(reduce_func=rfunc, metric_dtype="object")
        plotDict = {"bins": np.arange(0, 12), "xlabel": "Season length (days)"}
        plotFunc = maf.SummaryHistogram()
        displayDict["caption"] = f"Plot of the season length per season in the {fieldname} DDF."
        displayDict["order"] = order
        bundle = maf.MetricBundle(
            metric,
            ptslicer,
            fieldsqls["all"],
            info_label=" ".join([fieldname, info_labels["all"]]),
            plot_dict=plotDict,
            display_dict=displayDict,
            plot_funcs=[plotFunc],
        )
        bundle_list.append(bundle)

        # Median season Length
        metric = maf.SeasonLengthMetric(
            metric_name=f"{fieldname} Median Season Length", reduce_func=np.median
        )
        displayDict["caption"] = f"Median season length in the {fieldname} DDF."
        bundle_list.append(
            maf.MetricBundle(
                metric,
                ptslicer,
                fieldsqls[f],
                info_label=info_labels["all"],
                run_name=run_name,
                summary_metrics=[maf.MeanMetric()],
                plot_funcs=[],
                display_dict=displayDict,
            )
        )

        # Cumulative distribution - only for DDF fields
        if fieldname.endswith("WFD"):
            pass
        else:
            displayDict["group"] = "Progress"
            displayDict["subgroup"] = ""
            displayDict[
                "caption"
            ] = f"Cumulative number of visits for the {fieldname.replace('DD:', '')} field."
            slicer = maf.UniSlicer()
            metric = maf.CumulativeMetric(metric_name=f"{fieldname} Cumulative NVisits")
            metricb = maf.MetricBundle(
                metric,
                slicer,
                fieldsqls["all"],
                info_label=info_labels["all"],
                plot_funcs=[maf.XyPlotter()],
                run_name=run_name,
                display_dict=displayDict,
            )
            metricb.summary_metrics = []
            bundle_list.append(metricb)

    for b in bundle_list:
        b.set_run_name(run_name)
    bundleDict = maf.make_bundles_dict_from_list(bundle_list)

    return bundleDict



__all__ = (
    "NChangesMetric",
    "MinTimeBetweenStatesMetric",
    "NStateChangesFasterThanMetric",
    "MaxStateChangesWithinMetric",
    "TeffMetric",
    "OpenShutterFractionMetric",
    "BruteOSFMetric",
)

import numpy as np

from .base_metric import BaseMetric


class NChangesMetric(BaseMetric):
    """Compute the number of times a column value changes.
    (useful for filter changes in particular).
    """

    def __init__(self, col="filter", order_by="observationStartMJD", **kwargs):
        self.col = col
        self.order_by = order_by
        super(NChangesMetric, self).__init__(col=[col, order_by], units="#", **kwargs)

    def run(self, data_slice, slice_point=None):
        idxs = np.argsort(data_slice[self.order_by])
        diff = data_slice[self.col][idxs][1:] != data_slice[self.col][idxs][:-1]
        return np.size(np.where(diff == True)[0])


class MinTimeBetweenStatesMetric(BaseMetric):
    """Compute the minimum time between changes of state in a column value.
    (useful for calculating fastest time between filter changes in particular).
    Returns delta time in minutes!

    Parameters
    ----------
    change_col : `str`
        Column that we are tracking changes in.
    time_col : str
        Column with the time of each visit
    """

    def __init__(
        self,
        change_col="filter",
        time_col="observationStartMJD",
        metric_name=None,
        **kwargs,
    ):
        self.change_col = change_col
        self.time_col = time_col
        if metric_name is None:
            metric_name = "Minimum time between %s changes (minutes)" % (change_col)
        super(MinTimeBetweenStatesMetric, self).__init__(
            col=[change_col, time_col], metric_name=metric_name, units="", **kwargs
        )

    def run(self, data_slice, slice_point=None):
        # Sort on time, to be sure we've got changes in the right order.
        idxs = np.argsort(data_slice[self.time_col])
        changes = data_slice[self.change_col][idxs][1:] != data_slice[self.change_col][idxs][:-1]
        condition = np.where(changes == True)[0]
        changetimes = data_slice[self.time_col][idxs][1:][condition]
        prevchangetime = np.concatenate(
            (
                np.array([data_slice[self.time_col][idxs][0]]),
                data_slice[self.time_col][idxs][1:][condition][:-1],
            )
        )
        dtimes = changetimes - prevchangetime
        dtimes *= 24 * 60
        if dtimes.size == 0:
            return self.badval
        return dtimes.min()


class NStateChangesFasterThanMetric(BaseMetric):
    """
    Compute the number of changes of state that happen faster than 'cutoff'.
    (useful for calculating time between filter changes in particular).

    Parameters
    ----------
    change_col : `str`
        Column that we are tracking changes in.
    time_col : str
        Column with the time of each visit
    cutoff : `float`
        The cutoff value for the time between changes (in minutes).
    """

    def __init__(
        self,
        change_col="filter",
        time_col="observationStartMJD",
        metric_name=None,
        cutoff=20,
        **kwargs,
    ):
        if metric_name is None:
            metric_name = "Number of %s changes faster than <%.1f minutes" % (
                change_col,
                cutoff,
            )
        self.change_col = change_col
        self.time_col = time_col
        # Convert cutoff from minutes to days.
        self.cutoff = cutoff / 24.0 / 60.0
        super(NStateChangesFasterThanMetric, self).__init__(
            col=[change_col, time_col], metric_name=metric_name, units="#", **kwargs
        )

    def run(self, data_slice, slice_point=None):
        # Sort on time, to be sure we've got changes in the right order.
        idxs = np.argsort(data_slice[self.time_col])
        changes = data_slice[self.change_col][idxs][1:] != data_slice[self.change_col][idxs][:-1]
        condition = np.where(changes == True)[0]
        changetimes = data_slice[self.time_col][idxs][1:][condition]
        prevchangetime = np.concatenate(
            (
                np.array([data_slice[self.time_col][idxs][0]]),
                data_slice[self.time_col][idxs][1:][condition][:-1],
            )
        )
        dtimes = changetimes - prevchangetime
        return np.where(dtimes < self.cutoff)[0].size


class MaxStateChangesWithinMetric(BaseMetric):
    """Compute the maximum number of changes of state that occur
    within a given timespan.
    (useful for calculating time between filter changes in particular).

    Parameters
    ----------
    change_col : `str`
        Column that we are tracking changes in.
    time_col : str
        Column with the time of each visit
    timespan : `float`
        The timespan to count the number of changes within (in minutes).
    """

    def __init__(
        self,
        change_col="filter",
        time_col="observationStartMJD",
        metric_name=None,
        timespan=20,
        **kwargs,
    ):
        if metric_name is None:
            metric_name = "Max number of %s changes within %.1f minutes" % (
                change_col,
                timespan,
            )
        self.change_col = change_col
        self.time_col = time_col
        self.timespan = timespan / 24.0 / 60.0  # Convert timespan from minutes to days.
        super(MaxStateChangesWithinMetric, self).__init__(
            col=[change_col, time_col], metric_name=metric_name, units="#", **kwargs
        )

    def run(self, data_slice, slice_point=None):
        # This operates slightly differently from the metrics above;
        # those calculate only successive times between changes, but here
        # we must calculate the actual times of each change.
        # Check if there was only one observation (and return 0 if so).
        if data_slice[self.change_col].size == 1:
            return 0
        # Sort on time, to be sure we've got changes in the right order.
        idxs = np.argsort(data_slice[self.time_col])
        changes = data_slice[self.change_col][idxs][:-1] != data_slice[self.change_col][idxs][1:]
        condition = np.where(changes == True)[0]
        changetimes = data_slice[self.time_col][idxs][1:][condition]
        # If there are 0 filter changes ...
        if changetimes.size == 0:
            return 0
        # Otherwise ..
        ct_plus = changetimes + self.timespan
        indx2 = np.searchsorted(changetimes, ct_plus, side="right")
        indx1 = np.arange(changetimes.size)
        nchanges = indx2 - indx1
        return nchanges.max()


class TeffMetric(BaseMetric):
    """
    Effective time equivalent for a given set of visits.
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        metric_name="tEff",
        fiducial_depth=None,
        teff_base=30.0,
        normed=False,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        if fiducial_depth is None:
            # From reference von Karman 500nm zenith seeing of 0.69"
            # median zenith dark seeing from sims_skybrightness_pre
            # airmass = 1
            # 2 "snaps" of 15 seconds each
            # m5_flat_sed sysEngVals from rubin_sim
            #   commit 6d03bd49550972e48648503ed60784a4e6775b82 (2021-05-18)
            # These include constants from:
            #   https://github.com/lsst-pst/syseng_throughputs/blob/master/notebooks/generate_sims_values.ipynb
            #   commit 7abb90951fcbc70d9c4d0c805c55a67224f9069f (2021-05-05)
            # See https://github.com/lsst-sims/smtn-002/blob/master/notebooks/teff_fiducial.ipynb
            self.depth = {
                "u": 23.71,
                "g": 24.67,
                "r": 24.24,
                "i": 23.82,
                "z": 23.21,
                "y": 22.40,
            }
        else:
            if isinstance(fiducial_depth, dict):
                self.depth = fiducial_depth
            else:
                raise ValueError("fiducial_depth should be None or dictionary")
        self.teff_base = teff_base
        self.normed = normed
        if self.normed:
            units = ""
        else:
            units = "seconds"
        super(TeffMetric, self).__init__(
            col=[m5_col, filter_col], metric_name=metric_name, units=units, **kwargs
        )
        if self.normed:
            self.comment = "Normalized effective time"
        else:
            self.comment = "Effect time"
        self.comment += " of a series of observations, evaluating the equivalent amount of time"
        self.comment += " each observation would require if taken at a fiducial limiting magnitude."
        self.comment += " Fiducial depths are : %s" % self.depth
        if self.normed:
            self.comment += " Normalized by the total amount of time actual on-sky."

    def run(self, data_slice, slice_point=None):
        filters = np.unique(data_slice[self.filter_col])
        teff = 0.0
        for f in filters:
            match = np.where(data_slice[self.filter_col] == f)[0]
            teff += (10.0 ** (0.8 * (data_slice[self.m5_col][match] - self.depth[f]))).sum()
        teff *= self.teff_base
        if self.normed:
            # Normalize by the t_eff equivalent if each observation
            # was at the fiducial depth.
            teff = teff / (self.teff_base * data_slice[self.m5_col].size)
        return teff


class OpenShutterFractionMetric(BaseMetric):
    """Compute the fraction of time the shutter is open
    compared to the total time spent observing.
    """

    def __init__(
        self,
        metric_name="OpenShutterFraction",
        slew_time_col="slewTime",
        exp_time_col="visitExposureTime",
        visit_time_col="visitTime",
        **kwargs,
    ):
        self.exp_time_col = exp_time_col
        self.visit_time_col = visit_time_col
        self.slew_time_col = slew_time_col
        super(OpenShutterFractionMetric, self).__init__(
            col=[self.exp_time_col, self.visit_time_col, self.slew_time_col],
            metric_name=metric_name,
            units="OpenShutter/TotalTime",
            **kwargs,
        )
        self.comment = "Open shutter time (%s total) divided by total visit time " "(%s) + slewtime (%s)." % (
            self.exp_time_col,
            self.visit_time_col,
            self.slew_time_col,
        )

    def run(self, data_slice, slice_point=None):
        result = np.sum(data_slice[self.exp_time_col]) / np.sum(
            data_slice[self.slew_time_col] + data_slice[self.visit_time_col]
        )
        return result


class BruteOSFMetric(BaseMetric):
    """Assume I can't trust the slewtime or visittime colums.
    This computes the fraction of time the shutter is open,
    with no penalty for the first exposure after a long gap
    (e.g., 1st exposure of the night).
    Presumably, the telescope will need to focus, so there's not much a
    scheduler could do to optimize keeping the shutter open after a closure.

    Parameters
    ----------
    maxgap : `float`
        The maximum gap between observations, in minutes.
        Assume anything longer the dome has closed.
    fudge : `float`
        Fudge factor if a constant has to be added to the exposure time values.
        This time (in seconds) is added to the exposure time.
    exp_time_col : `str`
        The name of the exposure time column. Assumed to be in seconds.
    mjd_col : `str`
        The name of the start of the exposures. Assumed to be in units of days.
    """

    def __init__(
        self,
        metric_name="BruteOSFMetric",
        exp_time_col="visitExposureTime",
        mjd_col="observationStartMJD",
        maxgap=10.0,
        fudge=0.0,
        **kwargs,
    ):
        self.exp_time_col = exp_time_col
        self.maxgap = maxgap / 60.0 / 24.0  # convert from min to days
        self.mjd_col = mjd_col
        self.fudge = fudge
        super(BruteOSFMetric, self).__init__(
            col=[self.exp_time_col, mjd_col],
            metric_name=metric_name,
            units="OpenShutter/TotalTime",
            **kwargs,
        )

    def run(self, data_slice, slice_point=None):
        times = np.sort(data_slice[self.mjd_col])
        diff = np.diff(times)
        good = np.where(diff < self.maxgap)
        open_time = np.sum(diff[good]) * 24.0 * 3600.0
        result = np.sum(data_slice[self.exp_time_col] + self.fudge) / float(open_time)
        return result
